#!/usr/bin/env python
#
# gllinevector_funcs.py - OpenGL 1.4 functions used by the GLLineVector class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>

"""This module provides functions which are used by the :class:`.GLLineVector`
class to render :class:`.Image` overlays as line vector images in an OpenGL 1.4
compatible manner.


This module uses functions in the :mod:`.gl14.glvector_funcs` module, which
contains logic used for rendering both ``GLRGBVector`` and ``GLLineVector``
instances.


A :class:`.GLLineVertices` instance is used to generate line vertices and
texture coordinates for each voxel in the image. A fragment shader (the same
as that used by the :class:`.GLRGBVector` class) is used to colour each line
according to the orientation of the underlying vector.
"""


from __future__ import division

import logging

import numpy                   as np

import OpenGL.GL               as gl

import fsl.transform.affine    as affine
import fsleyes.gl.gllinevector as gllinevector
import fsleyes.gl.resources    as glresources
from . import                     glvector_funcs


log = logging.getLogger(__name__)


def init(self):
    """Compiles and configures the vertex/fragment shader programs, generates
    line vertices, and adds some listeners to properties of the
    :class:`.LineVectorOpts` instance associated with the vector
    :class:`.Image` overlay. This involves calls to the :func:`compileShaders`,
    :func:`updateShaderState`, and :func:`updateVertices` functions.
    """

    self.shader       = None
    self.lineVertices = None

    self._vertexResourceName = '{}_{}_vertices'.format(
        type(self).__name__, id(self.vectorImage))

    compileShaders(   self)
    updateShaderState(self)
    updateVertices(   self)

    opts = self.opts

    def vertexUpdate(*a):
        updateVertices(self)
        updateShaderState(self)
        self.notify()

    name = '{}_vertices'.format(self.name)

    opts.addListener('transform',   name, vertexUpdate, weak=False)
    opts.addListener('directed',    name, vertexUpdate, weak=False)
    opts.addListener('unitLength',  name, vertexUpdate, weak=False)
    opts.addListener('lengthScale', name, vertexUpdate, weak=False)
    opts.addListener('orientFlip',  name, vertexUpdate, weak=False)


def destroy(self):
    """Destroys the vertex/fragment shader programs and the
    ``GLLineVertices`` instance, and removes property listeners from the
    :class:`.LineVectorOpts` instance.
    """

    glvector_funcs.destroy(self)

    name = '{}_vertices'.format(self.name)

    self.opts.removeListener('transform',   name)
    self.opts.removeListener('directed',    name)
    self.opts.removeListener('unitLength',  name)
    self.opts.removeListener('lengthScale', name)
    self.opts.removeListener('orientFlip',  name)

    glresources.delete(self._vertexResourceName)


def compileShaders(self):
    """Compiles shader programs via the
    :func:`.gl14.glvector_funcs.compileShaders` function,
    and calls the :func:`updateVertices` function.
    """
    glvector_funcs.compileShaders(self, 'gllinevector')
    updateVertices(self)


def updateVertices(self):
    """Creates/refreshes the :class:`.GLLineVertices` instance which is used to
    generate line vertices and texture coordinates. If the ``GLLineVertices``
    instance exists and is up to date (see the
    :meth:`.GLLineVertices.calculateHash` method), this function does nothing.
    """

    if self.lineVertices is None:
        self.lineVertices = glresources.get(
            self._vertexResourceName, gllinevector.GLLineVertices, self)

    if hash(self.lineVertices) != self.lineVertices.calculateHash(self):

        log.debug('Re-generating line vertices '
                  'for {}'.format(self.vectorImage))

        self.lineVertices.refresh(self)
        glresources.set(self._vertexResourceName,
                        self.lineVertices,
                        overwrite=True)


def updateShaderState(self):
    """Updates all fragment/vertex shader program variables.  The fragment
    shader is configured by the
    :func:`.gl21.glvector_funcs.updateFragmentShaderState` function.
    """

    image = self.vectorImage

    glvector_funcs.updateShaderState(self)

    shape    = list(image.shape[:3])
    invShape = [1.0 / s for s in shape] + [0]

    self.shader.load()

    self.shader.setVertParam('invImageShape', invShape)

    self.shader.unload()

    return True


def preDraw(self, xform=None, bbox=None):
    """Initialises the GL state ready for drawing the :class:`.GLLineVector`.
    """
    gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
    self.shader.load()


def draw2D(self, zpos, axes, xform=None, bbox=None):
    """Draws the line vertices corresponding to a 2D plane located
    at the specified Z location.
    """

    opts                = self.opts
    vertices, voxCoords = self.lineVertices.getVertices2D(self,
                                                          zpos,
                                                          axes,
                                                          bbox=bbox)

    if vertices.size == 0:
        return

    self.shader.setAtt('voxCoord', voxCoords)
    self.shader.loadAtts()

    v2d = opts.getTransform('voxel', 'display')

    if xform is None: xform = v2d
    else:             xform = affine.concat(xform, v2d)

    gl.glPushMatrix()
    gl.glMultMatrixf(np.array(xform, dtype=np.float32).ravel('F'))

    gl.glVertexPointer(3, gl.GL_FLOAT, 0, vertices)

    gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
    gl.glLineWidth(opts.lineWidth)
    gl.glDrawArrays(gl.GL_LINES, 0, vertices.size // 3)

    gl.glPopMatrix()


def draw3D(self, xform=None, bbox=None):
    pass


def drawAll(self, axes, zposes, xforms):
    """Draws line vertices corresponding to each Z location. """
    for zpos, xform in zip(zposes, xforms):
        draw2D(self, zpos, axes, xform)


def postDraw(self, xform=None, bbox=None):
    """Clears the GL state after drawing the :class:`.GLLineVector`. """
    self.shader.unload()
    gl.glDisableClientState(gl.GL_VERTEX_ARRAY)



#!/usr/bin/env python
#
# timeseries.py - DataSeries classes used by the TimeSeriesPanel.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides a number of :class:`.DataSeries` sub-classes which
are use by the :class:`.TimeSeriesPanel`. The following classes are provided:

.. autosummary::
   :nosignatures:

   VoxelTimeSeries
   ComplexTimeSeries
   ImaginaryTimeSeries
   MagnitudeTimeSeries
   PhaseTimeSeries
   FEATTimeSeries
   FEATPartialFitTimeSeries
   FEATEVTimeSeries
   FEATResidualTimeSeries
   FEATModelFitTimeSeries
   MelodicTimeSeries
   MeshTimeSeries
"""


import numpy as np

import fsl.utils.deprecated as deprecated
import fsleyes_props        as props
import fsleyes.strings      as strings
import fsleyes.colourmaps   as fslcm
from . import                  dataseries


class VoxelTimeSeries(dataseries.VoxelDataSeries):
    """A ``VoxelTimeSeries`` is a ``VoxelDataSeries`` which represents
    time series data.
    """


class ComplexTimeSeries(VoxelTimeSeries):
    """A :class:`VoxelTimeSeries` to display time series from 4D
    complex images. The :meth:`getData` method returns the real component
    of the data..


    The :meth:`extraSeries` method returns additional series based on
    the values of the :attr:`plotImaginary`, :attr:`plotMagnitude` and
    :attr:`plotPhase` properties. The :meth:`extraSeries` method will
    return instances of the following classes:

    .. autosummary::
       :nosignatures:

       ImaginaryTimeSeries
       MagnitudeTimeSeries
       PhaseTimeSeries
    """


    plotReal = props.Boolean(default=True)
    """If ``True``, the :meth:`getData` method will return the real
    component time series data.
    """


    plotImaginary = props.Boolean(default=False)
    """If ``True``, the :meth:`extraSeries` method will return an
    :class:`ImaginaryTimeSeries` instance, containing the imaginary
    component data.
    """


    plotMagnitude = props.Boolean(default=False)
    """If ``True``, the :meth:`extraSeries` method will return a
    :class:`MagnitudeTimeSeries` instance, containing the complex
    magnitude.
    """


    plotPhase = props.Boolean(default=False)
    """If ``True``, the :meth:`extraSeries` method will return a
    :class:`PhaseTimeSeries` instance, containing the complex
    phase.
    """


    def __init__(self, overlay, overlayList, displayCtx, plotPanel):
        """Create a ``ComplexTimeSeries``. All arguments are passed
        through to the :class:`VoxelTimeSeries` constructor.
        """

        VoxelTimeSeries.__init__(
            self, overlay, overlayList, displayCtx, plotPanel)

        self.__imagts = ImaginaryTimeSeries(
            overlay, overlayList, displayCtx, plotPanel)
        self.__magts = MagnitudeTimeSeries(
            overlay, overlayList, displayCtx, plotPanel)
        self.__phasets = PhaseTimeSeries(
            overlay, overlayList, displayCtx, plotPanel)

        for ts in (self.__imagts, self.__magts, self.__phasets):
            ts.colour = fslcm.randomDarkColour()
            ts.bindProps('alpha',     self)
            ts.bindProps('lineWidth', self)
            ts.bindProps('lineStyle', self)


    def makeLabel(self):
        """Returns a string representation of this ``ComplexTimeSeries``
        instance.
        """
        return '{} ({})'.format(VoxelTimeSeries.makeLabel(self),
                                strings.labels[self])


    def getData(self):
        """If :attr:`plotReal` is true, returns the real component
        of the complex data. Otherwise returns ``(None, None)``.
        """
        if not self.plotReal:
            return None, None
        return VoxelTimeSeries.getData(self)


    def extraSeries(self):
        """Returns a list of additional series to be plotted, based
        on the values of the :attr:`plotImaginary`, :attr:`plotMagnitude`
        and :attr:`plotPhase` properties.
        """

        extras = []
        if self.plotImaginary: extras.append(self.__imagts)
        if self.plotMagnitude: extras.append(self.__magts)
        if self.plotPhase:     extras.append(self.__phasets)
        return extras


    def dataAtCurrentVoxel(self):
        """Returns the real component of the data at the current voxel. """
        data = VoxelTimeSeries.dataAtCurrentVoxel(self)
        if data is not None:
            data = data.real
        return data


class ImaginaryTimeSeries(VoxelTimeSeries):
    """An ``ImaginaryTimeSeries`` represents the imaginary component
    of a complex-valued image. ``ImaginaryTimeSeries`` instances
    are created by :class:`ComplexTimeSeries` instances.
    """


    def makeLabel(self):
        """Returns a string representation of this ``ImaginaryTimeSeries``
        instance.
        """
        return '{} ({})'.format(VoxelTimeSeries.makeLabel(self),
                                strings.labels[self])


    def dataAtCurrentVoxel(self):
        """Returns the imaginary component of the data at the current voxel.
        """
        data = VoxelTimeSeries.dataAtCurrentVoxel(self)
        if data is not None:
            data = data.imag
        return data


class MagnitudeTimeSeries(VoxelTimeSeries):
    """An ``MagnitudeTimeSeries`` represents the magnitude of a complex-valued
    image. ``MagnitudeTimeSeries`` instances are created by
    :class:`ComplexTimeSeries` instances.
    """


    def makeLabel(self):
        """Returns a string representation of this ``MagnitudeTimeSeries``
        instance.
        """
        return '{} ({})'.format(VoxelTimeSeries.makeLabel(self),
                                strings.labels[self])


    def dataAtCurrentVoxel(self):
        """Returns the magnitude of the data at the current voxel. """
        data = VoxelTimeSeries.dataAtCurrentVoxel(self)
        if data is not None:
            real = data.real
            imag = data.imag
            data = np.sqrt(real ** 2 + imag ** 2)
        return data


class PhaseTimeSeries(VoxelTimeSeries):
    """An ``PhaseTimeSeries`` represents the phase of a complex-valued
    image. ``PhaseTimeSeries`` instances are created by
    :class:`ComplexTimeSeries` instances.
    """


    def makeLabel(self):
        """Returns a string representation of this ``PhaseTimeSeries``
        instance.
        """
        return '{} ({})'.format(VoxelTimeSeries.makeLabel(self),
                                strings.labels[self])


    def dataAtCurrentVoxel(self):
        """Returns the phase of the data at the current voxel. """
        data = VoxelTimeSeries.dataAtCurrentVoxel(self)
        if data is not None:
            real = data.real
            imag = data.imag
            data = np.arctan2(imag, real)
        return data


class FEATTimeSeries(VoxelTimeSeries):
    """A :class:`VoxelTimeSeries` class for use with :class:`FEATImage`
    instances, containing some extra FEAT specific options.


    The ``FEATTimeSeries`` class acts as a container for several
    ``TimeSeries`` instances, each of which represent some part of a FEAT
    analysis. The data returned by a call to :meth:`.getData` on a
    ``FEATTimeSeries`` instance returns the fMRI time series data
    (``filtered_func_data`` in the ``.feat`` directory).


    The :meth:`extraSeries` method may be used to retrieve a list of all the
    other ``TimeSeries`` instances which are associated with the
    ``FEATTimeSeries`` instance - all of these ``DataSeries`` instances, in
    addition to this ``FEATTimeSeries`` instasnce, should be plotted.


    For example, if the :attr:`plotData` and :attr:`plotFullModelFit` settings
    are ``True``, the :meth:`extraSeries` method will return a list containing
    one ``TimeSeries`` instance, containing the full model fit, for the voxel
    in question.


    The following classes are used to represent the various parts of a FEAT
    analysis:

    .. autosummary::
       :nosignatures:

       FEATEVTimeSeries
       FEATResidualTimeSeries
       FEATPartialFitTimeSeries
       FEATModelFitTimeSeries
    """


    plotData = props.Boolean(default=True)
    """If ``True``, the FEAT input data is plotted. """


    plotFullModelFit = props.Boolean(default=True)
    """If ``True``, the FEAT full model fit is plotted. """


    plotResiduals = props.Boolean(default=False)
    """If ``True``, the FEAT model residuals are plotted. """


    plotEVs = props.List(props.Boolean(default=False))
    """A list of ``Boolean`` properties, one for each EV in the FEAT analysis.
    For elements that are ``True``, the corresponding FEAT EV time course is
    plotted.
    """


    plotPEFits = props.List(props.Boolean(default=False))
    """A list of ``Boolean`` properties, one for each EV in the FEAT analysis.
    For elements that are ``True``, the model fit for the corresponding FEAT
    EV is plotted.
    """


    plotCOPEFits = props.List(props.Boolean(default=False))
    """A list of ``Boolean`` properties, one for each EV in the FEAT analysis.
    For elements that are ``True``, the model fit for the corresponding FEAT
    contrast is plotted.
    """


    plotPartial = props.Choice()
    """Plot the raw data, after regression against a chosen EV or contrast.
    The options are populated in the :meth:`__init__` method.
    """


    def __init__(self, *args, **kwargs):
        """Create a ``FEATTimeSeries``.

        All arguments are passed through to the :class:`VoxelTimeSeries`
        constructor.
        """

        VoxelTimeSeries.__init__(self, *args, **kwargs)

        numEVs    = self.overlay.numEVs()
        numCOPEs  = self.overlay.numContrasts()
        copeNames = self.overlay.contrastNames()

        reduceOpts = ['none'] + \
                     ['PE{}'.format(i + 1) for i in range(numEVs)]

        for i in range(numCOPEs):
            name = 'COPE{} ({})'.format(i + 1, copeNames[i])
            reduceOpts.append(name)

        self.getProp('plotPartial').setChoices(reduceOpts, instance=self)

        for i in range(numEVs):
            self.plotPEFits.append(False)
            self.plotEVs   .append(False)

        for i in range(numCOPEs):
            self.plotCOPEFits.append(False)

        self.__fullModelTs =  None
        self.__partialTs   =  None
        self.__resTs       =  None
        self.__evTs        = [None] * numEVs
        self.__peTs        = [None] * numEVs
        self.__copeTs      = [None] * numCOPEs

        if not self.overlay.hasStats():
            self.plotFullModelFit = False

        self.addListener('plotFullModelFit',
                         self.name,
                         self.__plotFullModelFitChanged)
        self.addListener('plotResiduals',
                         self.name,
                         self.__plotResidualsChanged)
        self.addListener('plotPartial',
                         self.name,
                         self.__plotPartialChanged)

        self.addListener('plotEVs',      self.name, self.__plotEVChanged)
        self.addListener('plotPEFits',   self.name, self.__plotPEFitChanged)
        self.addListener('plotCOPEFits', self.name, self.__plotCOPEFitChanged)

        # plotFullModelFit defaults to True, so
        # force the model fit ts creation here
        self.__plotFullModelFitChanged()


    def getData(self):
        """Returns the fMRI time series data at the current voxel. Or,
        if :attr:`plotData` is ``False``, returns ``(None, None)``.
        """
        if not self.plotData:
            return None, None
        return VoxelTimeSeries.getData(self)


    def extraSeries(self):
        """Returns a list containing all of the ``TimeSeries`` instances
        which should be plotted in place of this ``FEATTimeSeries``.
        """

        modelts = []

        if self.plotFullModelFit:      modelts.append(self.__fullModelTs)
        if self.plotResiduals:         modelts.append(self.__resTs)
        if self.plotPartial != 'none': modelts.append(self.__partialTs)

        for i in range(self.overlay.numEVs()):
            if self.plotPEFits[i]:
                modelts.append(self.__peTs[i])

        for i in range(self.overlay.numEVs()):
            if self.plotEVs[i]:
                modelts.append(self.__evTs[i])

        for i in range(self.overlay.numContrasts()):
            if self.plotCOPEFits[i]:
                modelts.append(self.__copeTs[i])

        return modelts


    @deprecated.deprecated('0.31.0', '1.0.0', 'Use extraSeries instead')
    def getModelTimeSeries(self):
        return self.extraSeries()


    def __getContrast(self, fitType, idx):
        """Returns a contrast vector for the given model fit type, and index.

        :arg fitType: either ``'full'``, ``'pe'``, or ``'cope'``. If
                      ``'full'``, the ``idx`` argument is ignored.

        :arg idx:     The EV or contrast index for ``'pe'`` or ``'cope'`` model
                      fits.
        """

        if fitType == 'full':
            return [1] * self.overlay.numEVs()
        elif fitType == 'pe':
            con      = [0] * self.overlay.numEVs()
            con[idx] = 1
            return con
        elif fitType == 'cope':
            return self.overlay.contrasts()[idx]


    def __createModelTs(self, tsType, *args, **kwargs):
        """Creates a ``TimeSeries`` instance of the given ``tsType``, and
        sets its display settings  according to those of this
        ``FEATTimeSeries``.

        :arg tsType: The type to create, e.g. :class:`FEATModelFitTimeSeries`,
                     :class:`FEATEVTimeSeries`, etc.

        :arg args:   Passed to the ``tsType`` constructor.

        :arg kwargs: Passed to the ``tsType`` constructor.
        """

        ts = tsType(self.overlay,
                    self.overlayList,
                    self.displayCtx,
                    self.plotPanel,
                    self,
                    *args,
                    **kwargs)

        ts.alpha     = self.alpha
        ts.lineWidth = self.lineWidth
        ts.lineStyle = self.lineStyle

        if isinstance(ts, FEATModelFitTimeSeries) and ts.fitType == 'full':
            ts.colour = (0, 0, 0.8)
        else:
            ts.colour = fslcm.randomDarkColour()

        return ts


    def __plotPartialChanged(self, *a):
        """Called when the :attr:`plotPartial` setting changes.

        If necessary, creates and caches a :class:`FEATPartialFitTimeSeries`
        instance.
        """

        partial = self.plotPartial

        if partial == 'none' and self.__partialTs is not None:
            self.__partialTs = None
            return

        partial = partial.split()[0]

        # fitType is either 'cope' or 'pe'
        fitType = partial[:-1].lower()
        idx     = int(partial[-1]) - 1

        self.__partialTs = self.__createModelTs(
            FEATPartialFitTimeSeries,
            self.__getContrast(fitType, idx),
            fitType,
            idx)


    def __plotResidualsChanged(self, *a):
        """Called when the :attr:`plotResiduals` setting changes.

        If necessary, creates and caches a :class:`FEATResidualTimeSeries`
        instance.
        """

        if not self.plotResiduals:
            self.__resTs = None
            return

        self.__resTs = self.__createModelTs(FEATResidualTimeSeries)


    def __plotEVChanged(self, *a):
        """Called when the :attr:`plotEVs` setting changes.

        If necessary, creates and caches one or more :class:`FEATEVTimeSeries`
        instances.
        """

        for evnum, plotEV in enumerate(self.plotEVs):

            if not self.plotEVs[evnum]:
                self.__evTs[evnum] = None

            elif self.__evTs[evnum] is None:
                self.__evTs[evnum] = self.__createModelTs(
                    FEATEVTimeSeries, evnum)


    def __plotCOPEFitChanged(self, *a):
        """Called when the :attr:`plotCOPEFits` setting changes.

        If necessary, creates and caches one or more
        :class:`FEATModelFitTimeSeries` instances.
        """

        for copenum, plotCOPE in enumerate(self.plotCOPEFits):

            if not self.plotCOPEFits[copenum]:
                self.__copeTs[copenum] = None

            elif self.__copeTs[copenum] is None:
                self.__copeTs[copenum] = self.__createModelTs(
                    FEATModelFitTimeSeries,
                    self.__getContrast('cope', copenum),
                    'cope',
                    copenum)


    def __plotPEFitChanged(self, *a):
        """Called when the :attr:`plotPEFits` setting changes.

        If necessary, creates and caches one or more
        :class:`FEATModelFitTimeSeries` instances.
        """

        for evnum, plotPE in enumerate(self.plotPEFits):

            if not self.plotPEFits[evnum]:
                self.__peTs[evnum] = None

            elif self.__peTs[evnum] is None:
                self.__peTs[evnum] = self.__createModelTs(
                    FEATModelFitTimeSeries,
                    self.__getContrast('pe', evnum),
                    'pe',
                    evnum)


    def __plotFullModelFitChanged(self, *a):
        """Called when the :attr:`plotFullModelFit` setting changes.

        If necessary, creates and caches a
        :class:`FEATModelFitTimeSeries` instance.
        """

        if not self.plotFullModelFit:
            self.__fullModelTs = None
            return

        self.__fullModelTs = self.__createModelTs(
            FEATModelFitTimeSeries, self.__getContrast('full', -1), 'full', -1)


class FEATPartialFitTimeSeries(VoxelTimeSeries):
    """A :class:`VoxelTimeSeries` class which represents the partial model
    fit of an EV or contrast from a FEAT analysis at a specific voxel.
    Instances of this class are created by the :class:`FEATTimeSeries` class.
    """
    def __init__(self,
                 overlay,
                 overlayList,
                 displayCtx,
                 plotPanel,
                 parentTs,
                 contrast,
                 fitType,
                 idx):
        """Create a ``FEATPartialFitTimeSeries``.

        :arg overlay:     The :class:`.FEATImage` instance to extract the data
                          from.

        :arg overlayList: The :class:`.OverlayList` instance.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg plotPanel:   The :class:`TimeSeriesPanel` which owns this
                          ``FEATPartialFitTimeSeries``.

        :arg parentTs:    The :class:`.FEATTimeSeries` instance that has
                          created this ``FEATPartialFitTimeSeries``.

        :arg contrast:    The contrast vector to calculate the partial model
                          fit for.

        :arg fitType:     The model fit type, either ``'full'``, ``'pe'`` or
                          ``'cope'``.

        :arg idx:         If the model fit type is ``'pe'`` or ``'cope'``,
                          the EV/contrast index.
        """
        VoxelTimeSeries.__init__(
            self, overlay, overlayList, displayCtx, plotPanel)

        self.parentTs = parentTs
        self.contrast = contrast
        self.fitType  = fitType
        self.idx      = idx


    def dataAtCurrentVoxel(self):
        """Returns the partial model fit for the voxel and model fit type
        specified in the constructop.

        See the :meth:`.FEATImage.partialFit` method.
        """
        opts   = self.displayCtx.getOpts(self.overlay)
        coords = opts.getVoxel()

        if coords is None:
            return None

        return self.overlay.partialFit(self.contrast, coords)


class FEATEVTimeSeries(dataseries.DataSeries):
    """A :class:`TimeSeries` class which represents the time course of an
    EV from a FEAT analysis. Instances of this class are created by the
    :class:`FEATTimeSeries` class.
    """

    def __init__(self,
                 overlay,
                 overlayList,
                 displayCtx,
                 plotPanel,
                 parentTs,
                 idx):
        """Create a ``FEATEVTimeSeries``.

        :arg overlay:     The :class:`.FEATImage` instance to extract the data
                          from.

        :arg overlayList: The :class:`.OverlayList` instance.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg plotPanel:   The :class:`TimeSeriesPanel` which owns this
                          ``FEATEVTimeSeries``.

        :arg parentTs:    The :class:`.FEATTimeSeries` instance that has
                          created this ``FEATEVTimeSeries``.

        :arg idx:         The EV index.
        """
        dataseries.DataSeries.__init__(
            self, overlay, overlayList, displayCtx, plotPanel)

        self.parentTs = parentTs
        self.idx      = idx


    def makeLabel(self):
        """Returns a string representation of this ``FEATEVTimeSeries``
        instance.
        """

        display = self.displayCtx.getDisplay(self.overlay)

        return '{} EV{} ({})'.format(
            display.name,
            self.idx + 1,
            self.overlay.evNames()[self.idx])


    def getData(self):
        """Returns the time course of the EV specified in the constructor. """

        opts   = self.displayCtx.getOpts(self.overlay)
        coords = opts.getVoxel()
        design = self.overlay.getDesign(coords)
        ydata  = design[:, self.idx]
        xdata  = np.arange(len(ydata))

        return xdata, ydata


class FEATResidualTimeSeries(VoxelTimeSeries):
    """A :class:`VoxelTimeSeries` class which represents the time course of
    the residuals from a FEAT analysis at a specific voxel. Instances of this
    class are created by the :class:`FEATTimeSeries` class.
    """

    def __init__(self, overlay, overlayList, displayCtx, plotPanel, parentTs):
        """Create a ``FEATResidualTimeSeries``.

        :arg overlay:     The :class:`.FEATImage` instance to extract the data
                          from.

        :arg overlayList: The :class:`.OverlayList` instance.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg plotPanel:   The :class:`TimeSeriesPanel` which owns this
                          ``FEATResidualTimeSeries``.

        :arg parentTs:    The :class:`.FEATTimeSeries` instance that has
                          created this ``FEATResidualTimeSeries``.
        """
        VoxelTimeSeries.__init__(
            self, overlay, overlayList, displayCtx, plotPanel)
        self.parentTs = parentTs


    def makeLabel(self):
        """Returns a string representation of this ``FEATResidualTimeSeries``
        instance.
        """
        return '{} ({})'.format(self.parentTs.makeLabel(),
                                strings.labels[self])


    def dataAtCurrentVoxel(self):
        """Returns the residuals for the current voxel. """

        opts  = self.displayCtx.getOpts(self.overlay)
        voxel = opts.getVoxel()

        if voxel is None:
            return None

        x, y, z = voxel
        data    = self.overlay.getResiduals()[x, y, z, :]

        return data


class FEATModelFitTimeSeries(VoxelTimeSeries):
    """A :class:`TimeSeries` class which represents the time course for
    a model fit from a FEAT analysis at a specific voxel. Instances of this
    class are created by the :class:`FEATTimeSeries` class.
    """

    def __init__(self,
                 overlay,
                 overlayList,
                 displayCtx,
                 plotPanel,
                 parentTs,
                 contrast,
                 fitType,
                 idx):
        """Create a ``FEATModelFitTimeSeries``.

        :arg overlay:     The :class:`.FEATImage` instance to extract the data
                          from.

        :arg overlayList: The :class:`.OverlayList` instance.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg plotPanel:   The :class:`TimeSeriesPanel` which owns this
                          ``FEATModelFitTimeSeries``.

        :arg parentTs:    The :class:`.FEATTimeSeries` instance that has
                          created this ``FEATModelFitTimeSeries``.

        :arg contrast:    The contrast vector to calculate the partial model
                          fit for.

        :arg fitType:     The model fit type, either ``'full'``, ``'pe'`` or
                          ``'cope'``.

        :arg idx:         If the model fit type is ``'pe'`` or ``'cope'``,
        """

        if fitType not in ('full', 'cope', 'pe'):
            raise ValueError('Unknown model fit type {}'.format(fitType))

        VoxelTimeSeries.__init__(
            self, overlay, overlayList, displayCtx, plotPanel)
        self.parentTs = parentTs
        self.fitType  = fitType
        self.idx      = idx
        self.contrast = contrast


    def makeLabel(self):
        """Returns a string representation of this ``FEATModelFitTimeSeries``
        instance.
        """

        label = '{} ({})'.format(
            self.parentTs.makeLabel(),
            strings.labels[self, self.fitType])

        if self.fitType == 'full':
            return label

        elif self.fitType == 'cope':
            return label.format(
                self.idx + 1,
                self.overlay.contrastNames()[self.idx])

        elif self.fitType == 'pe':
            return label.format(self.idx + 1)


    def dataAtCurrentVoxel(self):
        """Returns the FEAT model fit at the current voxel. """

        opts     = self.displayCtx.getOpts(self.overlay)
        voxel    = opts.getVoxel()
        contrast = self.contrast

        if voxel is None:
            return None

        return self.overlay.fit(contrast, voxel)


class MelodicTimeSeries(dataseries.DataSeries):
    """A :class:`.DataSeries` class which encapsulates the time course for
    one component of a :class:`.MelodicImage`. The :meth:`getData` method
    returns the time course of the component specified by the current
    :class:`.NiftiOpts.volume`.
    """

    def __init__(self, overlay, overlayList, displayCtx, plotPanel):
        """Create a ``MelodicTimeSeries``.

        :arg overlay:     A :class:`.MelodicImage` overlay.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg plotPanel:   The :class:`TimeSeriesPanel` which owns this
                          ``MelodicTimeSeries``.
        """
        dataseries.DataSeries.__init__(
            self, overlay, overlayList, displayCtx, plotPanel)


    def getComponent(self):
        """Returns the index (starting from 0) of the current Melodic
        component, as dictated by the :class:`.NiftiOpts.volume` property.
        """
        opts = self.displayCtx.getOpts(self.overlay)
        return opts.volume


    def makeLabel(self):
        """Returns a string representation of this ``MelodicTimeSeries``. """

        display = self.displayCtx.getDisplay(self.overlay)
        return '{} [component {}]'.format(display.name,
                                          self.getComponent() + 1)


    def getData(self):
        """Returns the time course of the current Melodic component. """

        component = self.getComponent()
        ydata     = self.overlay.getComponentTimeSeries(component)
        xdata     = np.arange(len(ydata))
        return xdata, ydata


class MeshTimeSeries(dataseries.DataSeries):
    """A ``MeshTimeSeries`` object encapsulates the time course for a
    :class:`.Mesh` overlay which has some time series vertex data
    associated with it. See the :attr:`.MeshOpts.vertexData` property.
    """


    def __init__(self, overlay, overlayList, displayCtx, plotPanel):
        """Create a ``MeshTimeSeries`` instance.

        :arg overlay:     The :class:`.Mesh` instance to extract the data from.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg plotPanel:   The :class:`TimeSeriesPanel` which owns this
                          ``TimeSeries``.
        """
        dataseries.DataSeries.__init__(
            self, overlay, overlayList, displayCtx, plotPanel)


    def makeLabel(self):
        """Returns a label to use for this ``MeshTimeSeries`` on the
        legend.
        """

        display = self.displayCtx.getDisplay(self.overlay)

        if self.__haveData():

            opts    = display.opts
            vidx    = opts.getVertex()

            return '{} [{}]'.format(display.name, vidx)

        else:
            return display.name


    def __haveData(self):
        """Returns ``True`` if there is currently time series data to show
        for this ``MeshTimeSeries``, ``False`` otherwise.
        """
        opts = self.displayCtx.getOpts(self.overlay)
        vidx = opts.getVertex()
        vd   = opts.getVertexData()

        return vidx is not None and vd is not None and vd.shape[1] > 1


    def getData(self):
        """Returns the data at the current location for the
        :class:`.Mesh`, or ``(None, None)`` if there is no data.
        """

        if not self.__haveData():
            return None, None

        opts = self.displayCtx.getOpts(self.overlay)
        vidx = opts.getVertex()
        vd   = opts.getVertexData()

        ydata = vd[vidx, :]
        xdata = np.arange(len(ydata))

        return xdata, ydata



#!/usr/bin/env python
#
# histogramseries.py - The HistogramSeries class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`HistogramSeries`,
 :class:`ImageHistogramSeries`, :class:`ComplexHistogramSeries`, and
 :class:`MeshHistogramSeries` classes, used by the :class:`.HistogramPanel`
 for plotting histogram data.

Two standalone functions are also defined in this module:

  .. autosummary::
     :nosignatures:

     histogram
     autoBin
"""

import logging

import numpy as np

import fsl.utils.cache              as cache
import fsleyes_widgets.utils.status as status
import fsleyes_props                as props
import fsleyes.colourmaps           as fslcm
from . import                          dataseries


log = logging.getLogger(__name__)


class HistogramSeries(dataseries.DataSeries):
    """A ``HistogramSeries`` generates histogram data from an overlay. It is
    the base class for the :class:`ImageHistogramSeriess` and
    :class:`MeshHistogramSeries` classes.
    """


    nbins = props.Int(minval=10, maxval=1000, default=100, clamped=False)
    """Number of bins to use in the histogram. This value is overridden
    by the :attr:`autoBin` setting.
    """


    autoBin = props.Boolean(default=True)
    """If ``True``, the number of bins used for each :class:`HistogramSeries`
    is calculated automatically. Otherwise, :attr:`HistogramSeries.nbins` bins
    are used.
    """


    ignoreZeros = props.Boolean(default=True)
    """If ``True``, zeros are excluded from the calculated histogram. """


    includeOutliers = props.Boolean(default=False)
    """If ``True``, values which are outside of the :attr:`dataRange` are
    included in the histogram end bins.
    """


    dataRange = props.Bounds(ndims=1, clamped=False)
    """Specifies the range of data which should be included in the histogram.
    See the :attr:`includeOutliers` property.
    """


    def __init__(self, overlay, overlayList, displayCtx, plotPanel):
        """Create a ``HistogramSeries``.

        :arg overlay:     The overlay from which the data to be plotted is
                          retrieved.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg plotPanel:   The :class:`.HistogramPanel` that owns this
                          ``HistogramSeries``.
        """

        log.debug('New HistogramSeries instance for {} '.format(overlay.name))

        dataseries.DataSeries.__init__(
            self, overlay, overlayList, displayCtx, plotPanel)

        self.__nvals              = 0
        self.__dataKey            = None
        self.__xdata              = np.array([])
        self.__ydata              = np.array([])
        self.__finiteData         = np.array([])
        self.__nonZeroData        = np.array([])
        self.__clippedFiniteData  = np.array([])
        self.__clippedNonZeroData = np.array([])
        self.__dataCache          = cache.Cache(maxsize=10)
        self.__histCache          = cache.Cache(maxsize=100)

        self.addListener('dataRange',       self.name, self.__dataRangeChanged)
        self.addListener('nbins',           self.name, self.__histPropsChanged)
        self.addListener('autoBin',         self.name, self.__histPropsChanged)
        self.addListener('ignoreZeros',     self.name, self.__histPropsChanged)
        self.addListener('includeOutliers', self.name, self.__histPropsChanged)


    def destroy(self):
        """This needs to be called when this ``HistogramSeries`` instance
        is no longer being used.
        """

        self.removeListener('nbins',           self.name)
        self.removeListener('ignoreZeros',     self.name)
        self.removeListener('includeOutliers', self.name)
        self.removeListener('dataRange',       self.name)
        self.removeListener('nbins',           self.name)

        self.__dataCache.clear()
        self.__histCache.clear()
        self.__dataCache          = None
        self.__histCache          = None
        self.__nvals              = 0
        self.__dataKey            = None
        self.__xdata              = None
        self.__ydata              = None
        self.__finiteData         = None
        self.__nonZeroData        = None
        self.__clippedFiniteData  = None
        self.__clippedNonZeroData = None
        dataseries.DataSeries.destroy(self)


    def setHistogramData(self, data, key):
        """Must be called by sub-classes whenever the underlying histogram data
        changes.

        :arg data: A ``numpy`` array containing the data that the histogram is
                   to be calculated on. Pass in ``None``  to indicate that
                   there is currently no histogram data.

        :arg key:  Something which identifies the ``data``, and can be used as
                   a ``dict`` key.
        """

        if data is None:
            self.__nvals              = 0
            self.__dataKey            = None
            self.__xdata              = np.array([])
            self.__ydata              = np.array([])
            self.__finiteData         = np.array([])
            self.__nonZeroData        = np.array([])
            self.__clippedFiniteData  = np.array([])
            self.__clippedNonZeroData = np.array([])

            # force the panel to refresh
            with props.skip(self, 'dataRange', self.name):
                self.propNotify('dataRange')
            return

        # We cache the following data, based
        # on the provided key, so they don't
        # need to be recalculated:
        #  - finite data
        #  - non-zero data
        #  - finite minimum
        #  - finite maximum
        #
        # The cache size is restricted (see its
        # creation in __init__) so we don't blow
        # out RAM
        cached = self.__dataCache.get(key, None)

        if cached is None:

            log.debug('New histogram data {} - extracting '
                      'finite/non-zero data'.format(key))

            finData = data[np.isfinite(data)]
            nzData  = finData[finData != 0]
            dmin    = finData.min()
            dmax    = finData.max()

            self.__dataCache.put(key, (finData, nzData, dmin, dmax))
        else:
            log.debug('Got histogram data {} from cache'.format(key))
            finData, nzData, dmin, dmax = cached

        # The upper bound on the dataRange
        # is exclusive, so we initialise it
        # to a bit more than the data max.
        dist = (dmax - dmin) / 10000.0

        with props.suppressAll(self):

            self.dataRange.xmin = dmin
            self.dataRange.xmax = dmax + dist
            self.dataRange.xlo  = dmin
            self.dataRange.xhi  = dmax + dist
            self.nbins          = autoBin(nzData, self.dataRange.x)

            self.__dataKey     = key
            self.__finiteData  = finData
            self.__nonZeroData = nzData

            self.__dataRangeChanged()

        with props.skip(self, 'dataRange', self.name):
            self.propNotify('dataRange')


    def onDataRangeChange(self):
        """May be implemented by sub-classes. Is called when the
        :attr:`dataRange` changes.
        """
        pass


    def getData(self):
        """Overrides :meth:`.DataSeries.getData`.

        Returns  a tuple containing the ``(x, y)`` histogram data.
        """

        return self.__xdata, self.__ydata


    @property
    def binWidth(self):
        """Returns the width of one bin for this :class:`HistogramSeries`. """
        lo, hi = self.dataRange.x
        return (hi - lo) / self.nbins


    def getVertexData(self):
        """Returns a ``numpy`` array of shape ``(N, 2)``, which contains a
        set of "vertices" which can be used to display the histogram data
        as a filled polygon.
        """

        x, y = self.getData()

        if x is None or y is None:
            return None

        verts = np.zeros((len(x) * 2, 2), dtype=x.dtype)

        verts[  :,   0] = x.repeat(2)
        verts[ 1:-1, 1] = y.repeat(2)

        return verts


    @property
    def numHistogramValues(self):
        """Returns the number of values which were used in calculating the
        histogram.
        """
        return self.__nvals


    def __dataRangeChanged(self, *args, **kwargs):
        """Called when the :attr:`dataRange` property changes, and also by the
        :meth:`__initProperties` and :meth:`__volumeChanged` methods.
        """

        finData = self.__finiteData
        nzData  = self.__nonZeroData

        self.__clippedFiniteData  = finData[(finData >= self.dataRange.xlo) &
                                            (finData <  self.dataRange.xhi)]
        self.__clippedNonZeroData = nzData[ (nzData  >= self.dataRange.xlo) &
                                            (nzData  <  self.dataRange.xhi)]

        self.onDataRangeChange()
        self.__histPropsChanged()


    def __histPropsChanged(self, *a):
        """Called internally, and when any histogram settings change.
        Re-calculates the histogram data.
        """

        log.debug('Calculating histogram for '
                  'overlay {}'.format(self.overlay.name))

        status.update('Calculating histogram for '
                      'overlay {}'.format(self.overlay.name))

        if np.isclose(self.dataRange.xhi, self.dataRange.xlo):
            self.__xdata = np.array([])
            self.__ydata = np.array([])
            self.__nvals = 0
            return

        if self.ignoreZeros:
            if self.includeOutliers: data = self.__nonZeroData
            else:                    data = self.__clippedNonZeroData
        else:
            if self.includeOutliers: data = self.__finiteData
            else:                    data = self.__clippedFiniteData

        # Figure out the number of bins to use
        if self.autoBin: nbins = autoBin(data, self.dataRange.x)
        else:            nbins = self.nbins

        # nbins is unclamped, but
        # we don't allow < 10
        if nbins < 10:
            nbins = 10

        # Update the nbins property
        with props.skip(self, 'nbins', self.name):
            self.nbins = nbins

        # We cache calculated bins and counts
        # for each combination of parameters,
        # as histogram calculation can take
        # time.
        hrange  = (self.dataRange.xlo,  self.dataRange.xhi)
        drange  = (self.dataRange.xmin, self.dataRange.xmax)
        histkey = (self.__dataKey,
                   self.includeOutliers,
                   self.ignoreZeros,
                   hrange,
                   drange,
                   self.nbins)
        cached  = self.__histCache.get(histkey, None)

        if cached is not None:
            histX, histY, nvals = cached
        else:
            histX, histY, nvals = histogram(data,
                                            self.nbins,
                                            hrange,
                                            drange,
                                            self.includeOutliers,
                                            True)
            self.__histCache.put(histkey, (histX, histY, nvals))

        self.__xdata = histX
        self.__ydata = histY
        self.__nvals = nvals

        status.update('Histogram for {} calculated.'.format(
            self.overlay.name))

        log.debug('Calculated histogram for overlay '
                  '{} (number of values: {}, number '
                  'of bins: {})'.format(
                      self.overlay.name,
                      self.__nvals,
                      self.nbins))


class ImageHistogramSeries(HistogramSeries):
    """An ``ImageHistogramSeries`` instance manages generation of histogram
    data for an :class:`.Image` overlay.
    """


    showOverlay = props.Boolean(default=False)
    """If ``True``, a mask :class:`.ProxyImage` overlay is added to the
    :class:`.OverlayList`, which highlights the voxels that have been
    included in the histogram. The mask image is managed by the
    :class:`.HistogramProfile` instance, which manages histogram plot
    interaction.
    """


    showOverlayRange = props.Bounds(ndims=1)
    """Data range to display with the :attr:`.showOverlay` mask. """


    def __init__(self, *args, **kwargs):
        """Create an ``ImageHistogramSeries``. All arguments are passed
        through to :meth:`HistogramSeries.__init__`.
        """

        HistogramSeries.__init__(self, *args, **kwargs)

        self.__display = self.displayCtx.getDisplay(self.overlay)
        self.__opts    = self.displayCtx.getOpts(   self.overlay)

        self.__display.addListener('overlayType',
                                   self.name,
                                   self.__overlayTypeChanged)
        self.__opts   .addListener('volume',
                                   self.name,
                                   self.__volumeChanged)
        self.__opts   .addListener('volumeDim',
                                   self.name,
                                   self.__volumeChanged)

        self.__volumeChanged()


    def destroy(self):
        """Must be called when this ``ImageHistogramSeries`` is no longer
        needed. Removes some property listeners, and calls
        :meth:`HistogramSeries.destroy`.
        """

        HistogramSeries.destroy(self)

        self.__display.removeListener('overlayType', self.name)
        self.__opts   .removeListener('volume',      self.name)
        self.__opts   .removeListener('volumeDim',   self.name)


    def redrawProperties(self):
        """Overrides :meth:`.DataSeries.redrawProperties`. The
        ``HistogramSeries`` data does not need to be re-plotted when the
        :attr:`showOverlay` or :attr:`showOverlayRange` properties change.
        """

        propNames = dataseries.DataSeries.redrawProperties(self)

        propNames.remove('showOverlay')
        propNames.remove('showOverlayRange')

        return propNames


    def onDataRangeChange(self):
        """Overrides :meth:`HistogramSeries.onDataRangeChange`. Makes sure
        that the :attr:`showOverlayRange` limits are synced to the
        :attr:`HistogramSeries.dataRange`.
        """
        with props.suppress(self, 'showOverlayRange', notify=True):

            dlo, dhi = self.dataRange.x
            dist     = (dhi - dlo) / 10000.0

            needsInit = np.all(np.isclose(self.showOverlayRange.x, [0, 0]))

            self.showOverlayRange.xmin = dlo - dist
            self.showOverlayRange.xmax = dhi + dist

            if needsInit or not self.showOverlay:
                self.showOverlayRange.xlo = dlo
                self.showOverlayRange.xhi = dhi
            else:
                self.showOverlayRange.xlo = max(dlo, self.showOverlayRange.xlo)
                self.showOverlayRange.xhi = min(dhi, self.showOverlayRange.xhi)


    def __volumeChanged(self, *args, **kwargs):
        """Called when the :attr:`volume` property changes, and also by the
        :meth:`__init__` method.

        Passes the data to the :meth:`HistogramSeries.setHistogramData` method.
        """

        opts    = self.__opts
        overlay = self.overlay
        volkey  = (opts.volumeDim, opts.volume)

        self.setHistogramData(overlay[opts.index()], volkey)


    def __overlayTypeChanged(self, *a):
        """Called when the :attr:`.Display.overlayType` changes. When this
        happens, the :class:`.DisplayOpts` instance associated with the
        overlay gets destroyed and recreated. This method de-registers
        and re-registers property listeners as needed.
        """
        oldOpts     = self.__opts
        newOpts     = self.displayCtx.getOpts(self.overlay)
        self.__opts = newOpts

        oldOpts.removeListener('volume',    self.name)
        oldOpts.removeListener('volumeDim', self.name)
        newOpts.addListener(   'volume',    self.name, self.__volumeChanged)
        newOpts.addListener(   'volumeDim', self.name, self.__volumeChanged)


class ComplexHistogramSeries(ImageHistogramSeries):
    """Thre ``ComplexHistogramSeries`` class is a specialisation of the
    :class:`ImageHistogramSeries` for images with a complex data type.

    See also the :class:`.ComplexTimeSeries` and
    :class:`.ComplexPowerSpectrumSeries` classes.
    """


    plotReal      = props.Boolean(default=True)
    plotImaginary = props.Boolean(default=False)
    plotMagnitude = props.Boolean(default=False)
    plotPhase     = props.Boolean(default=False)


    def __init__(self, *args, **kwargs):
        """Create a ``ComplexHistogramSeries``. All arguments are passed
        through to the ``ImageHistogramSeries`` constructor.
        """
        ImageHistogramSeries.__init__(self, *args, **kwargs)

        self.__imaghs  = ImaginaryHistogramSeries(*args, **kwargs)
        self.__maghs   = MagnitudeHistogramSeries(*args, **kwargs)
        self.__phasehs = PhaseHistogramSeries(    *args, **kwargs)

        for hs in (self.__imaghs, self.__maghs, self.__phasehs):
            hs.colour = fslcm.randomDarkColour()
            hs.bindProps('alpha',           self)
            hs.bindProps('lineWidth',       self)
            hs.bindProps('lineStyle',       self)
            hs.bindProps('autoBin',         self)
            hs.bindProps('ignoreZeros',     self)
            hs.bindProps('includeOutliers', self)


    def extraSeries(self):
        """Returns a list containing an :class:`ImaginaryHistogramSeries`,
        :class:`MagnitudeHistogramSeries`, and/or
        :class:`PhaseHistogramSeries`, depending on the values of the
        :attr:`plotImaginary`, :attr:`plotMagnitude`, and :attr:`plotPhase`
        properties.
        """
        extras = []
        if self.plotImaginary: extras.append(self.__imaghs)
        if self.plotMagnitude: extras.append(self.__maghs)
        if self.plotPhase:     extras.append(self.__phasehs)
        return extras


    def getData(self):
        """Overrides :meth:`HistogramSeries.setHistogramData`. If
        :attr:`plotReal` is ``False``, returns ``(None, None)``. Otherwise
        returns the parent class implementation.
        """
        if self.plotReal: return ImageHistogramSeries.getData(self)
        else:             return None, None


    def setHistogramData(self, data, key):
        """Overrides :meth:`HistogramSeries.setHistogramData`.  The real
        component of the data is passed to the parent class implementation.
        """
        data = data.real
        ImageHistogramSeries.setHistogramData(self, data, key)


class ImaginaryHistogramSeries(ImageHistogramSeries):
    """Class which plots the histogram of the imaginary component
    of a complex-valued image.
    """
    def setHistogramData(self, data, key):
        data = data.imag
        ImageHistogramSeries.setHistogramData(self, data, key)


class MagnitudeHistogramSeries(ImageHistogramSeries):
    """Class which plots the histogram of the magnitude of a complex-valued
    image.
    """
    def setHistogramData(self, data, key):
        data = (data.real ** 2 + data.imag ** 2) ** 0.5
        ImageHistogramSeries.setHistogramData(self, data, key)


class PhaseHistogramSeries(ImageHistogramSeries):
    """Class which plots the histogram of the phase of a complex-valued image.
    """
    def setHistogramData(self, data, key):
        data = np.arctan2(data.imag, data.real)
        ImageHistogramSeries.setHistogramData(self, data, key)


class MeshHistogramSeries(HistogramSeries):
    """A ``MeshHistogramSeries`` instance manages generation of histogram
    data for a :class:`.Mesh` overlay.
    """

    def __init__(self, *args, **kwargs):
        """Create a ``MeshHistogramSeries``. All arguments are passed
        through to :meth:`HistogramSeries.__init__`.
        """

        HistogramSeries.__init__(self, *args, **kwargs)

        self.__opts = self.displayCtx.getOpts(self.overlay)

        self.__opts.addListener('vertexData',
                                self.name,
                                self.__vertexDataChanged)
        self.__opts.addListener('vertexDataIndex',
                                self.name,
                                self.__vertexDataChanged)

        self.__vertexDataChanged()


    def destroy(self):
        """Must be called when this ``MeshHistogramSeries`` is no longer
        needed. Calls :meth:`HistogramSeries.destroy` and removes some
        property listeners.
        """
        HistogramSeries.destroy(self)
        self.__opts.removeListener('vertexData',      self.name)
        self.__opts.removeListener('vertexDataIndex', self.name)
        self.__opts = None


    def __vertexDataChanged(self, *a):
        """Called when the :attr:`.MeshOpts.vertexData` or
        :attr:`.MeshOpts.vertexDataIndex` properties change. Updates the
        histogram data via :meth:`HistogramSeries.setHistogramData`.
        """

        vdname = self.__opts.vertexData
        vdi    = self.__opts.vertexDataIndex
        vd     = self.__opts.getVertexData()

        if vd is None: self.setHistogramData(None, None)
        else:          self.setHistogramData(vd[:, vdi],   (vdname, vdi))


def histogram(data,
              nbins,
              histRange,
              dataRange,
              includeOutliers=False,
              count=True):
    """Calculates a histogram of the given ``data``.

    :arg data:            The data to calculate a histogram foe

    :arg nbins:           Number of bins to use

    :arg histRange:       Tuple containing  the ``(low, high)`` data range
                          that the histogram is to be calculated on.

    :arg dataRange:       Tuple containing  the ``(min, max)`` range of
                          values in the data

    :arg includeOutliers: If ``True``, the outermost bins will contain counts
                          for values which are outside the ``histRange``.
                          Defaults to ``False``.

    :arg count:           If ``True`` (the default), the raw histogram counts
                          are returned. Otherwise they are converted into
                          probabilities.

    :returns:             A tuple containing:

                           - The ``x`` histogram data (bin edges)

                           - The ``y`` histogram data

                           - The total number of values that were used
                             in the histogram calculation
    """

    hlo, hhi = histRange
    dlo, dhi = dataRange

    # Calculate bin edges
    bins = np.linspace(hlo, hhi, nbins + 1)

    if includeOutliers:
        bins[ 0] = dlo
        bins[-1] = dhi

    # Calculate the histogram
    histX    = bins
    histY, _ = np.histogram(data.flat, bins=bins)
    nvals    = histY.sum()

    if not count:
        histY = histY / nvals

    return histX, histY, nvals


def autoBin(data, dataRange):
    """Calculates the number of bins which should be used for a histogram
    of the given data. The calculation is identical to that implemented
    in the original FSLView.

    :arg data:      The data that the histogram is to be calculated on.

    :arg dataRange: A tuple containing the ``(min, max)`` histogram range.
    """

    dMin, dMax = dataRange
    dRange     = dMax - dMin

    if np.isclose(dRange, 0):
        return 1

    binSize = np.power(10, np.ceil(np.log10(dRange) - 1) - 1)

    nbins = dRange / binSize

    while nbins < 100:
        binSize /= 2
        nbins    = dRange / binSize

    if issubclass(data.dtype.type, np.integer):
        binSize = max(1, np.ceil(binSize))

    adjMin = np.floor(dMin / binSize) * binSize
    adjMax = np.ceil( dMax / binSize) * binSize

    nbins = int((adjMax - adjMin) / binSize) + 1

    return nbins



#!/usr/bin/env python
#
# scene3dtoolbar.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import wx


import fsleyes_props                 as props
import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.toolbar               as fsltoolbar
import fsleyes.strings               as strings
import fsleyes.actions               as actions
import fsleyes.icons                 as fslicons
import fsleyes.tooltips              as fsltooltips



class Scene3DToolBar(ctrlpanel.ControlToolBar):
    """
    """


    showCursorAndLegend = props.Boolean(default=True)
    """This property is linked to a button on the toolbar which allows the
    user to simultaneously toggle the :attr:`.SceneOpts.showCursor` and
    :attr:`.Scene3DOpts.showLegend` properties.
    """


    def __init__(self, parent, overlayList, displayCtx, frame, panel):
        """Create a ``Scene3DToolBar``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        :arg panel:       The :class:`.Scene3DPanel` instance.
        """

        ctrlpanel.ControlToolBar.__init__(self,
                                          parent,
                                          overlayList,
                                          displayCtx,
                                          frame,
                                          height=24,
                                          kbFocus=True)

        self.panel = panel

        self.addListener('showCursorAndLegend',
                         self.name,
                         self.__showCursorAndLegendChanged)

        self.__makeTools()


    def destroy(self):
        """Must be called when this ``Scene3DToolBar`` is no longer in use.
        Removes some property listeners, and calls the base class
        implementation.
        """
        self.removeListener('showCursorAndLegend', self.name)

        ctrlpanel.ControlToolBar.destroy(self)


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``Scene3DToolBar`` is only intended to be added to
        :class:`.Scene3DPanel` views.
        """
        from fsleyes.views.scene3dpanel import Scene3DPanel
        return [Scene3DPanel]


    def __makeTools(self):
        """Called by :meth:`__init__`. Creates the toolbar widgets. """

        panel   = self.panel
        opts    = panel.sceneOpts
        profile = panel.getCurrentProfile()

        icons = {
            'screenshot'          : fslicons.findImageFile('camera24'),
            'resetDisplay'        : fslicons.findImageFile('resetZoom24'),
            'showCursorAndLegend' : [
                fslicons.findImageFile('addHighlight24'),
                fslicons.findImageFile('add24')],
            'movieMode'        : [
                fslicons.findImageFile('movieHighlight24'),
                fslicons.findImageFile('movie24')],
            'toggleCanvasSettingsPanel' : [
                fslicons.findImageFile('spannerHighlight24'),
                fslicons.findImageFile('spanner24')],
        }

        tooltips = {
            'screenshot'   : fsltooltips.actions[   panel,     'screenshot'],
            'resetDisplay' : fsltooltips.actions[   profile,   'resetDisplay'],
            'movieMode'    : fsltooltips.properties[panel,     'movieMode'],
            'showCursorAndLegend' : fsltooltips.properties[
                self, 'showCursorAndLegend'],
            'zoom'         : fsltooltips.properties[opts, 'zoom'],
            'toggleCanvasSettingsPanel' : fsltooltips.actions[
                panel, 'toggleCanvasSettingsPanel'],
        }

        targets = {
            'screenshot'                : panel,
            'resetDisplay'              : profile,
            'movieMode'                 : panel,
            'showCursorAndLegend'       : self,
            'zoom'                      : opts,
            'toggleCanvasSettingsPanel' : panel,
        }

        toolSpecs = [
            actions.ToggleActionButton(
                'toggleCanvasSettingsPanel',
                actionKwargs={'floatPane' : True},
                icon=icons['toggleCanvasSettingsPanel'],
                tooltip=tooltips['toggleCanvasSettingsPanel']),
            actions.ActionButton('screenshot',
                                 icon=icons['screenshot'],
                                 tooltip=tooltips['screenshot']),
            'div',
            props  .Widget(      'movieMode',
                                 icon=icons['movieMode'],
                                 tooltip=tooltips['movieMode']),
            props  .Widget(      'showCursorAndLegend',
                                 icon=icons['showCursorAndLegend'],
                                 tooltip=tooltips['showCursorAndLegend']),
            actions.ActionButton('resetDisplay',
                                 icon=icons['resetDisplay'],
                                 tooltip=tooltips['resetDisplay']),
            props.Widget(        'zoom',
                                 spin=True,
                                 slider=True,
                                 showLimits=False,
                                 spinWidth=5,
                                 tooltip=tooltips['zoom']),
        ]

        tools = []
        nav   = []

        for spec in toolSpecs:

            if spec == 'div':
                tools.append(fsltoolbar.ToolBarDivider(self,
                                                       height=24,
                                                       orient=wx.VERTICAL))
                continue

            widget    = props.buildGUI(self, targets[spec.key], spec)
            navWidget = widget

            if spec.key in ('zoom', ):
                widget = self.MakeLabelledTool(
                    widget,
                    strings.properties[targets[spec.key], spec.key])

            tools.append(widget)
            nav  .append(navWidget)

        self.SetTools(tools, destroy=True)
        self.setNavOrder(nav)


    def __showCursorAndLegendChanged(self, *a):
        """Called when the :attr:`showCursorAndLegend` property is changed.
        Propagates the change on to the :attr:`.Scene3DOpts.showCursor` and
        :attr:`.Scene3DOpts.showLegend` properties.
        """

        opts            = self.panel.sceneOpts
        opts.showCursor = self.showCursorAndLegend
        opts.showLegend = self.showCursorAndLegend



#!/usr/bin/env python
#
# histogrampanel.py - The HistogramPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`HistogramPanel`, which is a *FSLeyes view*
that plots the histogram of data from :class:`.Image` overlays.
"""


import logging

import wx

import numpy as np

import fsl.data.image                         as fslimage
import fsl.data.mesh                          as fslmesh
import fsleyes_props                          as props

import fsleyes.actions                        as actions
import fsleyes.actions.addroihistogram        as roihistogram
import fsleyes.overlay                        as fsloverlay
import fsleyes.plotting.histogramseries       as histogramseries
import fsleyes.controls.histogramcontrolpanel as histogramcontrolpanel
import fsleyes.controls.histogramtoolbar      as histogramtoolbar
from . import                                    plotpanel


log = logging.getLogger(__name__)


class HistogramPanel(plotpanel.OverlayPlotPanel):
    """An :class:`.OverlayPlotPanel` which plots histograms from
    :class:`.Image`     overlay data. A ``HistogramPanel`` looks something
    like this:

    .. image:: images/histogrampanel.png
       :scale: 50%
       :align: center


    A ``HistogramPanel`` plots one or more :class:`HistogramSeries` instances,
    each of which encapsulate histogram data from an :class:`.Image` overlay.


    A couple of control panels may be shown on a ``HistogramPanel``:

    .. autosummary::
       :nosignatures:

       ~fsleyes.controls.plotlistpanel.PlotListPanel
       ~fsleyes.controls.histogramcontrolpanel.HistogramControlPanel


    The following :mod:`actions` are provided, in addition to those already
    provided by the :class:`.PlotPanel`:

    .. autosummary::
       :nosignatures:

       toggleHistogramToolBar
       toggleHistogramControl

    Some tools are also available, to do various things:

    .. autosummary::
       :nosignatures:

       addROIHistogram
    """


    histType = props.Choice(('probability', 'count'))
    """The histogram type:

    =============== ==========================================================
    ``count``       The y axis represents the absolute number of values within
                    each bin
    ``probability`` The y axis represents the number of values within each
                    bin, divided by the total number of values.
    =============== ==========================================================
    """


    plotType = props.Choice(('centre', 'edge'))
    """How histograms are plotted:

    ========== ==========================================================
    ``centre`` Plot one data point at the centre of each bin
    ``edge``   Plot one data point at each bin edge - this produces a
               "stepped" plot.
    ========== ==========================================================
    """


    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create a ``HistogramPanel``.

        :arg parent:      The :mod:`wx` parent.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        """

        plotpanel.OverlayPlotPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        self.__currentHs = None

        self       .addListener('histType', self.name, self.draw)
        self       .addListener('plotType', self.name, self.draw)
        overlayList.addListener('overlays',
                                self.name,
                                self.__selectedOverlayChanged)
        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__selectedOverlayChanged)

        self.__roiHistAction = roihistogram.AddROIHistogramAction(
            overlayList,
            displayCtx,
            self)

        self.addROIHistogram.bindProps('enabled', self.__roiHistAction)

        self.initProfile()
        self.__selectedOverlayChanged()


    def destroy(self):
        """Removes some property listeners, and calls
        :meth:`.PlotPanel.destroy`.
        """

        self.__roiHistAction.destroy()

        self.__currentHs     = None
        self.__roiHistAction = None

        self            .removeListener('histType',        self.name)
        self            .removeListener('plotType',        self.name)
        self.overlayList.removeListener('overlays',        self.name)
        self.displayCtx .removeListener('selectedOverlay', self.name)

        plotpanel.OverlayPlotPanel.destroy(self)


    @actions.toggleControlAction(histogramcontrolpanel.HistogramControlPanel)
    def toggleHistogramControl(self, floatPane=False):
        """Shows/hides a :class:`.HistogramControlPanel`. See
        :meth:`.ViewPanel.togglePanel`.
        """
        self.togglePanel(histogramcontrolpanel.HistogramControlPanel,
                         self,
                         location=wx.RIGHT,
                         floatPane=floatPane)


    @actions.toggleControlAction(histogramtoolbar.HistogramToolBar)
    def toggleHistogramToolBar(self):
        """Shows/hides a :class:`.HistogramToolBar`. See
        :meth:`.ViewPanel.togglePanel`.
        """
        self.togglePanel(histogramtoolbar.HistogramToolBar, histPanel=self)


    @actions.toggleAction
    def toggleHistogramOverlay(self):
        """Toggles the value of the :attr:`.HistogramSeries.showOverlay`
        for the currently selected overlay (if possible).
        """
        # This action gets configured in the
        # __selectedOverlayChanged method
        pass


    @actions.action
    def addROIHistogram(self):
        """Runs an :class:`.AddROIHistogramAction`. """

        self.__roiHistAction()


    def getActions(self):
        """Overrides :meth:`.ActionProvider.getActions`. Returns all of the
        :mod:`.actions` that are defined on this ``HistogramPanel``.
        """
        actions = [self.screenshot,
                   self.importDataSeries,
                   self.exportDataSeries,
                   None,
                   self.toggleOverlayList,
                   self.togglePlotList,
                   self.toggleHistogramToolBar,
                   self.toggleHistogramControl,
                   self.toggleHistogramOverlay]

        names = [a.__name__ if a is not None else None for a in actions]

        return list(zip(names, actions))


    def getTools(self):
        """Returns a list of tools to be added to the ``FSLeyesFrame`` for
        ``HistogramPanel`` views.
        """
        return [self.addROIHistogram]


    def draw(self, *a):
        """Overrides :meth:`.PlotPanel.draw`. Passes some
        :class:`.HistogramSeries` instances to the
        :meth:`.PlotPanel.drawDataSeries` method.
        """

        if not self or self.destroyed:
            return

        hss = self.getDataSeriesToPlot()

        for hs in hss:
            with props.suppress(hs, 'label'):
                hs.label = self.displayCtx.getDisplay(hs.overlay).name

        if self.smooth or self.plotType == 'centre':
            self.drawDataSeries(hss)

        # use a step plot when plotting bin edges
        else:
            self.drawDataSeries(hss, drawstyle='steps-pre')

        self.drawArtists()


    def createDataSeries(self, overlay):
        """Creates a :class:`.HistogramSeries` instance for the specified
        overlay.
        """

        if isinstance(overlay, fsloverlay.ProxyImage):
            overlay = overlay.getBase()

        if isinstance(overlay, fslimage.Image):
            if overlay.iscomplex:
                hsType = histogramseries.ComplexHistogramSeries
            else:
                hsType = histogramseries.ImageHistogramSeries
        elif isinstance(overlay, fslmesh.Mesh):
            hsType = histogramseries.MeshHistogramSeries
        else:
            return None, None, None

        hs = hsType(overlay, self.overlayList, self.displayCtx, self)
        hs.colour      = self.getOverlayPlotColour(overlay)
        hs.lineStyle   = self.getOverlayPlotStyle(overlay)
        hs.lineWidth   = 2
        hs.alpha       = 1

        return hs, [], []


    def prepareDataSeries(self, hs):
        """Overrides :meth:`.PlotPanel.prepareDataSeries`.

        Performs some pre-processing on the data contained in the given
        :class:`.HistogramSeries` instance.
        """

        xdata, ydata = hs.getData()

        if xdata is None   or \
           ydata is None   or \
           len(xdata) == 0 or \
           len(ydata) == 0:
            return None, None

        # If smoothing is enabled, we just
        # need to provide the data as-is
        if self.smooth:
            xdata = np.array(xdata[:-1], dtype=np.float32)
            ydata = np.array(ydata,      dtype=np.float32)

        # If plotting on bin edges, we need to
        # munge the histogram data a bit so that
        # plt.plot(drawstyle='steps-pre') plots
        # it nicely.
        elif self.plotType == 'edge':
            xpad = np.zeros(len(xdata) + 1, dtype=np.float32)
            ypad = np.zeros(len(ydata) + 2, dtype=np.float32)

            xpad[ :-1] = xdata
            xpad[  -1] = xdata[-1]
            ypad[1:-1] = ydata
            xdata      = xpad
            ydata      = ypad

        # otherwise if plotting bin centres,
        # we need to offset the data, as it
        # is on bin edges
        elif self.plotType == 'centre':
            ydata = np.array(ydata,      dtype=np.float32)
            xdata = np.array(xdata[:-1], dtype=np.float32)
            xdata = xdata + 0.5 * (xdata[1] - xdata[0])

        # The passed-in series may just
        # be a DataSeries instance.
        if not isinstance(hs, histogramseries.HistogramSeries):
            return xdata, ydata

        # Or a HistogramSeries instance
        if self.histType == 'count':
            return xdata, ydata

        # Normalise by bin width to produce a
        # probability density function, so it
        # will look approximately the same,
        # regardless of the current nbins setting
        elif self.histType == 'probability':
            nvals = hs.numHistogramValues
            return xdata, ydata / (nvals * hs.binWidth)


    def __selectedOverlayChanged(self, *a):
        """Called when the :class:`.OverlayList` or the
        :attr:`.DisplayContext.selectedOverlay` changes. Configures the
        :meth:`toggleHistogramOverlay` action.
        """

        overlay = self.displayCtx.getSelectedOverlay()
        oldHs   = self.__currentHs
        newHs   = self.getDataSeries(overlay)
        enable  = (overlay is not None) and \
                  (newHs   is not None) and \
                  isinstance(overlay, fslimage.Image)

        self.toggleHistogramOverlay.enabled = enable

        if not enable or (oldHs is newHs):
            return

        self.__currentHs = newHs

        if oldHs is not None:
            self.toggleHistogramOverlay.unbindProps(
                'toggled', oldHs, 'showOverlay')

        self.toggleHistogramOverlay.bindProps(
            'toggled', newHs, 'showOverlay')



#!/usr/bin/env python
#
# orthoeditsettingspanel.py - The OrthoEditSettingsPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OrthoEditSettingsPanel`, a *FSLeyes*
control panel which contains settings to manage an
:class:`.OrthoEditProfile`.
"""

import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes_props                 as props
import fsleyes.strings               as strings
import fsleyes.tooltips              as fsltooltips


class OrthoEditSettingsPanel(ctrlpanel.SettingsPanel):
    """The ``OrthoEditSettingsPanel`` is a *FSLeyes* control panel which
    displays widgets allowing the user to adjust properties of an
    :class:`.OrthoEditProfile`.
    """


    def __init__(self, parent, overlayList, displayCtx, frame, ortho):
        """Create an ``OrthoEditSettingsPanel``.

        :arg parent:      The :mod:`wx` parent object
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        :arg ortho:       The :class:`.OrthoPanel` instance.
        """
        ctrlpanel.SettingsPanel.__init__(self,
                                         parent,
                                         overlayList,
                                         displayCtx,
                                         frame,
                                         kbFocus=True)

        self.__ortho = ortho
        self.__makeTools()


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``OrthoEditSettingsPanel`` is only intended to be added to
        :class:`.OrthoPanel` views.
        """
        from fsleyes.views.orthopanel import OrthoPanel
        return [OrthoPanel]


    def __makeTools(self):
        """Generates all of the tools and adds them to the :class:`.WidgetList`
        (see :class:`.SettingsPanel`),
        """

        ortho   = self.__ortho
        widgets = self.getWidgetList()
        profile = ortho.getCurrentProfile()

        if ortho.profile != 'edit':
            return

        generalProps = [
            props.Widget('mode',
                         labels=strings.choices[profile, 'mode'],
                         fixChoices=['nav', 'sel', 'desel', 'fill', 'selint']),
            props.Widget('drawMode'),
            props.Widget('selectionIs3D'),
            props.Widget('selectionSize',
                         slider=False,
                         spin=True,
                         showLimits=False),
            props.Widget('fillValue',  slider=False, spin=True),
            props.Widget('eraseValue', slider=False, spin=True),

            # 'showSelectionOutline',

            props.Widget('selectionCursorColour'),
            props.Widget('selectionOverlayColour'),
            props.Widget(
                'locationFollowsMouse',
                dependencies=['drawMode'],
                enabledWhen=lambda p, m: not m),
            props.Widget(
                'showSelection',
                dependencies=['drawMode'],
                enabledWhen=lambda p, m: not m)
        ]

        selintProps = [

            props.Widget(
                'localFill',
                dependencies=['mode'],
                enabledWhen=lambda p, m: m == 'selint'),
            props.Widget(
                'limitToRadius',
                dependencies=['mode'],
                enabledWhen=lambda p, m: m == 'selint'),

            props.Widget(
                'intensityThresLimit',
                spin=True,
                slider=False,
                showLimits=False,
                dependencies=['mode'],
                enabledWhen=lambda p, m: m == 'selint'),

            props.Widget(
                'intensityThres',
                spin=True,
                slider=True,
                showLimits=False,
                dependencies=['mode'],
                enabledWhen=lambda p, m: m == 'selint'),

            props.Widget(
                'searchRadius',
                spin=True,
                slider=True,
                showLimits=False,
                dependencies=['mode', 'limitToRadius'],
                enabledWhen=lambda p, m, r: m == 'selint' and r),
        ]

        widgets.AddGroup('general', strings.labels[self, 'general'])
        widgets.AddGroup('selint',  strings.labels[self, 'selint'])

        allWidgets = []

        for spec in generalProps:

            widget = props.buildGUI(widgets, profile, spec)
            allWidgets.append(widget)
            widgets.AddWidget(
                widget,
                displayName=strings.properties[profile, spec.key],
                tooltip=fsltooltips.properties[profile, spec.key],
                groupName='general')

        for spec in selintProps:

            widget = props.buildGUI(widgets, profile, spec)
            allWidgets.append(widget)
            widgets.AddWidget(
                widget,
                displayName=strings.properties[profile, spec.key],
                tooltip=fsltooltips.properties[profile, spec.key],
                groupName='selint')

        self.setNavOrder(allWidgets)
        widgets.Expand('general')
        widgets.Expand('selint')



#!/usr/bin/env python
#
# scene3dcanvas.py - The Scene3DCanvas class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`.Scene3DCanvas` class, which is used by
FSLeyes for its 3D view.
"""


import logging

import numpy     as np
import OpenGL.GL as gl

import fsl.data.mesh        as fslmesh
import fsl.data.image       as fslimage
import fsl.utils.idle       as idle
import fsl.transform.affine as affine

import fsleyes.gl.routines               as glroutines
import fsleyes.gl.globject               as globject
import fsleyes.displaycontext            as fsldisplay
import fsleyes.displaycontext.canvasopts as canvasopts


log = logging.getLogger(__name__)


class Scene3DCanvas(object):


    def __init__(self, overlayList, displayCtx):

        self.__name           = '{}_{}'.format(type(self).__name__, id(self))
        self.__opts           = canvasopts.Scene3DCanvasOpts()
        self.__overlayList    = overlayList
        self.__displayCtx     = displayCtx
        self.__viewMat        = np.eye(4)
        self.__projMat        = np.eye(4)
        self.__invViewProjMat = np.eye(4)
        self.__viewport       = None
        self.__resetLightPos  = True
        self.__glObjects      = {}

        overlayList.addListener('overlays',
                                self.__name,
                                self.__overlayListChanged)
        displayCtx.addListener('bounds',
                               self.__name,
                               self.__displayBoundsChanged)

        opts = self.opts
        opts.addListener('pos',          self.__name, self.Refresh)
        opts.addListener('showCursor',   self.__name, self.Refresh)
        opts.addListener('cursorColour', self.__name, self.Refresh)
        opts.addListener('bgColour',     self.__name, self.Refresh)
        opts.addListener('showLegend',   self.__name, self.Refresh)
        opts.addListener('occlusion',    self.__name, self.Refresh)
        opts.addListener('zoom',         self.__name, self.Refresh)
        opts.addListener('offset',       self.__name, self.Refresh)
        opts.addListener('rotation',     self.__name, self.Refresh)
        opts.addListener('highDpi',      self.__name, self.__highDpiChanged)


    def destroy(self):
        """
        """
        self.__overlayList.removeListener('overlays', self.__name)
        self.__displayCtx .removeListener('bounds',   self.__name)

        for ovl in list(self.__glObjects.keys()):
            self.__deregisterOverlay(ovl)

        self.__opts        = None
        self.__displayCtx  = None
        self.__overlayList = None
        self.__glObjects   = None


    @property
    def destroyed(self):
        """
        """
        return self.__overlayList is None


    @property
    def opts(self):
        """Returns a reference to the :class:`.Scene3DCanvasOpts` instance.
        """
        return self.__opts

    @property
    def resetLightPos(self):
        """By default, the :attr:`lightPos` is updated whenever the
        :attr:`.DisplayContext.bounds` change. This flag can be used to
        disable this behaviour.
        """
        return self.__resetLightPos


    @resetLightPos.setter
    def resetLightPos(self, reset):
        """Control whether the :attr:`lightPos` property is reset whenever
        the :attr:`.DisplayContext.bounds` change.
        """
        self.__resetLightPos = reset


    def defaultLightPos(self):
        """Resets the :attr:`lightPos` property to a sensible value. """
        b      = self.__displayCtx.bounds
        centre = np.array([b.xlo + 0.5 * (b.xhi - b.xlo),
                           b.ylo + 0.5 * (b.yhi - b.ylo),
                           b.zlo + 0.5 * (b.zhi - b.zlo)])
        self.opts.lightPos = centre + [b.xlen, b.ylen, 0]


    @property
    def viewMatrix(self):
        """Returns the view matrix for the current scene - this is an affine
        matrix which encodes the current :attr:`.Scene3DCanvasOpts.offset`,
        :attr:`.Scene3DCanvasOpts.zoom`,
        :attr:`.Scene3DCanvasOpts.rotation` and camera settings.

        See :meth:`__genViewMatrix`.
        """
        return self.__viewMat


    @property
    def viewScale(self):
        """Returns an affine matrix which encodes the current
        :attr:`.Scene3DCanvasOpts.zoom` setting.
        """
        return self.__viewScale


    @property
    def viewOffset(self):
        """Returns an affine matrix which encodes the current
        :attr:`.Scene3DCanvasOpts.offset` setting.
        """
        return self.__viewOffset


    @property
    def viewRotation(self):
        """Returns an affine matrix which encodes the current
        :attr:`.Scene3DCanvasOpts.rotation` setting.
        """
        return self.__viewRotate


    @property
    def viewCamera(self):
        """Returns an affine matrix which encodes the current camera.
        transformation. The initial camera orientation in the view shown by a
        :class:`Scene3DCanvas` is located on the positive Y axis, is oriented
        towards the positive Z axis, and is pointing towards the centre of the
        :attr:`.DisplayContext.displayBounds`.
        """
        return self.__viewCamera


    @property
    def projectionMatrix(self):
        """Returns the projection matrix. This is an affine matrix which
        converts from normalised device coordinates (NDCs, coordinates between
        -1 and +1) into viewport coordinates. The initial viewport for a
        :class:`Scene3DCanvas` is configured by the :func:`.routines.ortho`
        function.

        See :meth:`__setViewport`.
        """
        return self.__projMat


    @property
    def invViewProjectionMatrix(self):
        """Returns the inverse of the model-view-projection matrix, the
        equivalent of:

        ``invert(projectionMatrix * viewMatrix)``
        """
        return self.__invViewProjMat


    @property
    def viewport(self):
        """Returns a list of three ``(min, max)`` tuples which specify the
        viewport limits of the currently displayed scene.
        """
        return self.__viewport


    def canvasToWorld(self, xpos, ypos, near=True):
        """Transform the given x/y canvas coordinates into the display
        coordinate system. The calculated coordinates will be located on
        the near clipping plane.

        :arg near: If ``True`` (the default), the returned coordinate will
                   be located on the near clipping plane. Otherwise, the
                   coordinate will be located on the far clipping plane.
        """

        width, height = self.GetSize()

        # Normalise pixels to [-1, 1]
        xp = -1 + 2.0 * xpos / width
        yp = -1 + 2.0 * ypos / height

        # We set the Z coord so the resulting
        # coordinates will be located on either
        # the  near or clipping planes.
        if near: pos = [xp, yp, -1]
        else:    pos = [xp, yp,  1]

        # The first step is to convert mouse
        # coordinates from [-1, 1] to viewport
        # coodinates via the inverse projection
        # matrix.

        # The second step is to transform from
        # viewport coords into model-view coords.
        # This is easy - transform by the inverse
        # MV matrix.

        # We perform both of these steps in one
        # by concatenating then inverting the
        # view/projection matrices. This is
        # calculated and cached for us in the
        # __setViewport method.
        pos = affine.transform(pos, self.__invViewProjMat)

        return pos


    def getGLObject(self, overlay):
        """Returns the :class:`.GLObject` associated with the given overlay,
        or ``None`` if there is not one.
        """
        return self.__glObjects.get(overlay, None)


    def getGLObjects(self):
        """Returns two lists:

         - A list of overlays to be drawn
         - A list of corresponding :class:`GLObject` instances

        The lists are in the order that they should be drawn.

        This method also creates ``GLObject`` instances for any overlays
        in the :class:`.OverlayList` that do not have one.
        """

        overlays = self.__displayCtx.getOrderedOverlays()

        surfs = [o for o in overlays if isinstance(o, fslmesh.Mesh)]
        vols  = [o for o in overlays if isinstance(o, fslimage.Image)]
        other = [o for o in overlays if o not in surfs and o not in vols]

        overlays = []
        globjs   = []

        # If occlusion is on, we draw all surfaces first,
        # so they are on the scene regardless of volume
        # opacity.
        #
        # If occlusion is off, we draw all volumes
        # (without depth testing) first, and draw all
        # surfaces (with depth testing) afterwards.
        # In this way, the surfaces will be occluded
        # by the last drawn volume. I figure that this
        # is better than being occluded by *all* volumes,
        # regardless of depth or camera orientation.
        #
        # The one downside to this is that if a
        # transparent volume is in front of a surface,
        # the surface won't be shown.
        #
        # The only way to overcome this would be to
        # sort by depth on every render which, given
        # the possibility of volume clipping planes,
        # is a bit too complicated for my liking.
        if self.opts.occlusion: ovlOrder = surfs + vols + other
        else:                   ovlOrder = vols + surfs + other

        for ovl in ovlOrder:
            globj = self.getGLObject(ovl)

            # If there is no GLObject for this
            # overlay, create one, but don't
            # add it to the list (as creation
            # is done asynchronously).
            if globj is None:
                self.__registerOverlay(ovl)

            # Otherwise, if the value for this
            # overlay evaluates to False, that
            # means that it has been scheduled
            # for creation, but is not ready
            # yet.
            elif globj:
                overlays.append(ovl)
                globjs  .append(globj)

        return overlays, globjs


    def _initGL(self):
        """Called when the canvas is ready to be drawn on. """
        self.__overlayListChanged()
        self.__displayBoundsChanged()


    def __overlayListChanged(self, *a):
        """Called when the :class:`.OverlayList` changes. Destroys/creates
        :class:`.GLObject` instances as necessary.
        """

        # Destroy any GL objects for overlays
        # which are no longer in the list
        for ovl, globj in list(self.__glObjects.items()):
            if ovl not in self.__overlayList:
                self.__deregisterOverlay(ovl)

        # Create GLObjects for any
        # newly added overlays
        for ovl in self.__overlayList:
            if ovl not in self.__glObjects:
                self.__registerOverlay(ovl)


    def __highDpiChanged(self, *a):
        """Called when the :attr:`.Scene3DCanvasOpts.highDpi` property
        changes. Calls the :meth:`.GLCanvasTarget.EnableHighDPI` method.
        """
        self.EnableHighDPI(self.opts.highDpi)


    def __displayBoundsChanged(self, *a):
        """Called when the :attr:`.DisplayContext.bounds` change. Resets
        the :attr:`.Scene3DCanvasOpts.lightPos` property.
        """

        if self.resetLightPos:
            self.defaultLightPos()

        self.Refresh()


    def __registerOverlay(self, overlay):
        """
        """

        if not isinstance(overlay, (fslmesh.Mesh, fslimage.Image)):
            return

        log.debug('Registering overlay {}'.format(overlay))

        display = self.__displayCtx.getDisplay(overlay)

        if not self.__genGLObject(overlay):
            return

        display.addListener('enabled', self.__name, self.Refresh)
        display.addListener('overlayType',
                            self.__name,
                            self.__overlayTypeChanged)


    def __deregisterOverlay(self, overlay):
        """
        """

        log.debug('Deregistering overlay {}'.format(overlay))

        try:
            display = self.__displayCtx.getDisplay(overlay)
            display.removeListener('overlayType', self.__name)
            display.removeListener('enabled',     self.__name)
        except fsldisplay.InvalidOverlayError:
            pass

        globj = self.__glObjects.pop(overlay, None)

        if globj is not None:
            globj.deregister(self.__name)
            globj.destroy()


    def __genGLObject(self, overlay):
        """
        """

        if overlay in self.__glObjects:
            return False

        display = self.__displayCtx.getDisplay(overlay)

        if display.overlayType not in ('volume', 'mesh'):
            return False

        self.__glObjects[overlay] = False

        def create():

            if not self or self.destroyed:
                return

            if overlay not in self.__glObjects:
                return

            if not self._setGLContext():
                self.__glObjects.pop(overlay)
                return

            log.debug('Creating GLObject for {}'.format(overlay))

            globj = globject.createGLObject(overlay,
                                            self.__overlayList,
                                            self.__displayCtx,
                                            self,
                                            True)

            if globj is not None:
                globj.register(self.__name, self.Refresh)
                self.__glObjects[overlay] = globj

        idle.idle(create)
        return True


    def __overlayTypeChanged(self, value, valid, display, name):
        """
        """

        overlay = display.overlay
        globj   = self.__glObjects.pop(overlay, None)

        if globj is not None:
            globj.deregister(self.__name)
            globj.destroy()

        self.__genGLObject(overlay)
        self.Refresh()


    def __genViewMatrix(self, w, h):
        """Generate and return a transformation matrix to be used as the
        model-view matrix. This includes applying the current :attr:`zoom`,
        :attr:`rotation` and :attr:`offset` settings, and configuring
        the camera. This method is called by :meth:`__setViewport`.

        :arg w: Canvas width in pixels
        :arg h: Canvas height in pixels
        """

        opts   = self.opts
        b      = self.__displayCtx.bounds
        centre = [b.xlo + 0.5 * b.xlen,
                  b.ylo + 0.5 * b.ylen,
                  b.zlo + 0.5 * b.zlen]

        # The MV matrix comprises (in this order):
        #
        #    - A rotation (the rotation property)
        #
        #    - Camera configuration. With no rotation, the
        #      camera will be looking towards the positive
        #      Y axis (i.e. +y is forwards), and oriented
        #      towards the positive Z axis (i.e. +z is up)
        #
        #    - A translation (the offset property)
        #    - A scaling (the zoom property)

        # Scaling and rotation matrices. Rotation
        # is always around the centre of the
        # displaycontext bounds (the bounding
        # box which contains all loaded overlays).
        scale  = opts.zoom / 100.0
        scale  = affine.scaleOffsetXform([scale] * 3, 0)
        rotate = affine.rotMatToAffine(opts.rotation, centre)

        # The offset property is defined in x/y
        # pixels, normalised to [-1, 1]. We need
        # to convert them into viewport space,
        # where the horizontal axis maps to
        # (-xhalf, xhalf), and the vertical axis
        # maps to (-yhalf, yhalf). See
        # gl.routines.ortho.
        offset     = np.array(opts.offset[:] + [0])
        xlen, ylen = glroutines.adjust(b.xlen, b.ylen, w, h)
        offset[0]  = xlen * offset[0] / 2
        offset[1]  = ylen * offset[1] / 2
        offset     = affine.scaleOffsetXform(1, offset)

        # And finally the camera.
        eye     = list(centre)
        eye[1] += 1
        up      = [0, 0, 1]
        camera  = glroutines.lookAt(eye, centre, up)

        # Order is very important!
        xform = affine.concat(offset, scale, camera, rotate)
        np.array(xform, dtype=np.float32)

        self.__viewOffset = offset
        self.__viewScale  = scale
        self.__viewRotate = rotate
        self.__viewCamera = camera
        self.__viewMat    = xform


    def __setViewport(self):
        """Called by :meth:`_draw`. Configures the viewport and calculates
        the model-view trasformation matrix.

        :returns: ``True`` if the viewport was successfully configured,
                  ``False`` otherwise.
        """

        width, height = self.GetScaledSize()
        b             = self.__displayCtx.bounds
        blo           = [b.xlo, b.ylo, b.zlo]
        bhi           = [b.xhi, b.yhi, b.zhi]
        zoom          = self.opts.zoom / 100.0

        if width == 0 or height == 0:
            return False

        # We allow one dimension to be
        # flat, so we can display 2D
        # meshes (e.g. flattened surfaces)
        if np.sum(np.isclose(blo, bhi)) > 1:
            return False

        # Generate the view and projection matrices
        self.__genViewMatrix(width, height)
        projmat, viewport     = glroutines.ortho(blo, bhi, width, height, zoom)
        self.__projMat        = projmat
        self.__viewport       = viewport
        self.__invViewProjMat = affine.concat(self.__projMat, self.__viewMat)
        self.__invViewProjMat = affine.invert(self.__invViewProjMat)

        gl.glViewport(0, 0, width, height)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadMatrixf(self.__projMat.ravel('F'))
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()

        return True


    def _draw(self):
        """Draws the scene to the canvas. """

        if not self._setGLContext():
            return

        opts = self.opts
        glroutines.clear(opts.bgColour)

        if not self.__setViewport():
            return

        overlays, globjs = self.getGLObjects()

        if len(overlays) == 0:
            return

        # If occlusion is on, we offset the
        # depth of each overlay so that, where
        # a depth collision occurs, overlays
        # which are higher in the list will get
        # drawn above (closer to the screen)
        # than lower ones.
        depthOffset = affine.scaleOffsetXform(1, [0, 0, 0.1])
        depthOffset = np.array(depthOffset,    dtype=np.float32, copy=False)
        xform       = np.array(self.__viewMat, dtype=np.float32, copy=False)

        for ovl, globj in zip(overlays, globjs):

            display = self.__displayCtx.getDisplay(ovl)

            if not globj.ready():
                continue
            if not display.enabled:
                continue

            if opts.occlusion:
                xform = affine.concat(depthOffset, xform)
            elif isinstance(ovl, fslimage.Image):
                gl.glClear(gl.GL_DEPTH_BUFFER_BIT)

            log.debug('Drawing {} [{}]'.format(ovl, globj))

            globj.preDraw( xform=xform)
            globj.draw3D(  xform=xform)
            globj.postDraw(xform=xform)

        if opts.showCursor:
            with glroutines.enabled((gl.GL_DEPTH_TEST)):
                self.__drawCursor()

        if opts.showLegend:
            self.__drawLegend()

        # Testing click-to-near/far clipping plane transformation
        if hasattr(self, 'points'):
            colours = [(1, 0, 0, 1), (0, 0, 1, 1)]
            gl.glPointSize(5)

            gl.glBegin(gl.GL_LINES)
            for i, p in enumerate(self.points):
                gl.glColor4f(*colours[i % 2])
                p = affine.transform(p, self.viewMatrix)
                gl.glVertex3f(*p)
            gl.glEnd()


    def __drawCursor(self):
        """Draws three lines at the current :attr:`.DisplayContext.location`.
        """

        opts = self.opts
        b    = self.__displayCtx.bounds
        pos  = opts.pos

        points = np.array([
            [pos.x, pos.y, b.zlo],
            [pos.x, pos.y, b.zhi],
            [pos.x, b.ylo, pos.z],
            [pos.x, b.yhi, pos.z],
            [b.xlo, pos.y, pos.z],
            [b.xhi, pos.y, pos.z],
        ], dtype=np.float32)
        points = affine.transform(points, self.__viewMat)
        gl.glLineWidth(1)

        r, g, b = opts.cursorColour[:3]

        gl.glColor4f(r, g, b, 1)
        gl.glBegin(gl.GL_LINES)
        for p in points:
            gl.glVertex3f(*p)
        gl.glEnd()


    def __drawLegend(self):
        """Draws a legend in the bottom left corner of the screen, showing
        anatomical orientation.
        """

        copts      = self.opts
        b          = self.__displayCtx.bounds
        w, h       = self.GetSize()
        xlen, ylen = glroutines.adjust(b.xlen, b.ylen, w, h)

        # A line for each axis
        vertices       = np.zeros((6, 3), dtype=np.float32)
        vertices[0, :] = [-1,  0,  0]
        vertices[1, :] = [ 1,  0,  0]
        vertices[2, :] = [ 0, -1,  0]
        vertices[3, :] = [ 0,  1,  0]
        vertices[4, :] = [ 0,  0, -1]
        vertices[5, :] = [ 0,  0,  1]

        # Each axis line is scaled to
        # 60 pixels, and the legend is
        # offset from the bottom-left
        # corner by twice this amount.
        scale      = [xlen * 30.0 / w] * 3
        offset     = [-0.5 * xlen + 2.0 * scale[0],
                      -0.5 * ylen + 2.0 * scale[1],
                       0]

        # Apply the current camera
        # angle and rotation settings
        # to the legend vertices. Offset
        # anatomical labels off each
        # axis line by a small amount.
        rotation   = affine.decompose(self.__viewMat)[2]
        xform      = affine.compose(scale, offset, rotation)
        labelPoses = affine.transform(vertices * 1.2, xform)
        vertices   = affine.transform(vertices,       xform)

        # Draw the legend lines
        gl.glDisable(gl.GL_DEPTH_TEST)
        gl.glColor3f(*copts.cursorColour[:3])
        gl.glLineWidth(2)
        gl.glBegin(gl.GL_LINES)
        gl.glVertex3f(*vertices[0])
        gl.glVertex3f(*vertices[1])
        gl.glVertex3f(*vertices[2])
        gl.glVertex3f(*vertices[3])
        gl.glVertex3f(*vertices[4])
        gl.glVertex3f(*vertices[5])
        gl.glEnd()


        # Figure out the anatomical
        # labels for each axis.
        overlay = self.__displayCtx.getSelectedOverlay()
        dopts   = self.__displayCtx.getOpts(overlay)
        labels  = dopts.getLabels()[0]

        # getLabels returns (xlo, ylo, zlo, xhi, yhi, zhi) -
        # - rearrange them to (xlo, xhi, ylo, yhi, zlo, zhi)
        labels = [labels[0],
                  labels[3],
                  labels[1],
                  labels[4],
                  labels[2],
                  labels[5]]

        canvas = np.array([w, h])
        view   = np.array([xlen, ylen])

        # Draw each label
        for i in range(6):

            # Calculate pixel x/y
            # location for this label
            xx, xy = canvas * (labelPoses[i, :2] + 0.5 * view) / view

            # Calculate the size of the label
            # in pixels, so we can centre the
            # label
            tw, th = glroutines.text2D(labels[i], (xx, xy), 10, (w, h),
                                       calcSize=True)

            # Draw the text
            xx -= 0.5 * tw
            xy -= 0.5 * th
            gl.glColor3f(*copts.legendColour[:3])
            glroutines.text2D(labels[i], (xx, xy), 10, (w, h))


    def __drawLight(self):

        opts      = self.opts
        lightPos  = np.array(opts.lightPos)
        lightPos *= (opts.zoom / 100.0)

        gl.glColor4f(1, 1, 1, 1)
        gl.glPointSize(10)
        gl.glBegin(gl.GL_POINTS)
        gl.glVertex3f(*lightPos)
        gl.glEnd()

        b = self.__displayCtx.bounds
        centre = np.array([b.xlo + 0.5 * (b.xhi - b.xlo),
                           b.ylo + 0.5 * (b.yhi - b.ylo),
                           b.zlo + 0.5 * (b.zhi - b.zlo)])

        centre = affine.transform(centre, self.__viewMat)

        gl.glColor4f(1, 0, 1, 1)
        gl.glBegin(gl.GL_LINES)
        gl.glVertex3f(*lightPos)
        gl.glVertex3f(*centre)
        gl.glEnd()


    def __drawBoundingBox(self):
        b = self.__displayCtx.bounds
        xlo, xhi = b.x
        ylo, yhi = b.y
        zlo, zhi = b.z
        xlo += 0.1
        xhi -= 0.1
        vertices = np.array([
            [xlo, ylo, zlo],
            [xlo, ylo, zhi],
            [xlo, yhi, zlo],
            [xlo, yhi, zhi],
            [xhi, ylo, zlo],
            [xhi, ylo, zhi],
            [xhi, yhi, zlo],
            [xhi, yhi, zhi],

            [xlo, ylo, zlo],
            [xlo, yhi, zlo],
            [xhi, ylo, zlo],
            [xhi, yhi, zlo],
            [xlo, ylo, zhi],
            [xlo, yhi, zhi],
            [xhi, ylo, zhi],
            [xhi, yhi, zhi],

            [xlo, ylo, zlo],
            [xhi, ylo, zlo],
            [xlo, ylo, zhi],
            [xhi, ylo, zhi],
            [xlo, yhi, zlo],
            [xhi, yhi, zlo],
            [xlo, yhi, zhi],
            [xhi, yhi, zhi],
        ])
        vertices = affine.transform(vertices, self.__viewMat)


        gl.glLineWidth(2)
        gl.glColor3f(0.5, 0, 0)
        gl.glBegin(gl.GL_LINES)
        for v in vertices:
            gl.glVertex3f(*v)
        gl.glEnd()



#!/usr/bin/env python
#
# frameactions.py - Top level actions
#
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains various top-level FSLeyes actions which are
monkey-patched into the :class:`.FSLeyesFrame` class.
"""


import            os
import os.path as op

import                      fsleyes
import fsleyes.actions   as actions
import fsleyes.strings   as strings
from   fsleyes.frame import FSLeyesFrame



def addViewPanel(self, vpType, **kwargs):
    """Function shared by the add*Panel functions below. """
    vp = self.addViewPanel(vpType, **kwargs)
    self.viewPanelDefaultLayout(vp)
    vp.SetFocus()
    return vp


def addOrthoPanel(self, *args, **kwargs):
    """Adds a new :class:`.OrthoPanel`."""
    from fsleyes.views.orthopanel import OrthoPanel
    addViewPanel(self, OrthoPanel)


def addLightBoxPanel(self, *args, **kwargs):
    """Adds a new :class:`.LightBoxPanel`."""
    from fsleyes.views.lightboxpanel import LightBoxPanel
    addViewPanel(self, LightBoxPanel)


def addScene3DPanel(self, *args, **kwargs):
    """Adds a new :class:`.Scene3DPanel`."""
    from fsleyes.views.scene3dpanel import Scene3DPanel
    addViewPanel(self, Scene3DPanel)


def addTimeSeriesPanel(self, *args, **kwargs):
    """Adds a new :class:`.TimeSeriesPanel`."""
    from fsleyes.views.timeseriespanel import TimeSeriesPanel
    addViewPanel(self, TimeSeriesPanel)


def addHistogramPanel(self, *args, **kwargs):
    """Adds a new :class:`.HistogramPanel`."""
    from fsleyes.views.histogrampanel import HistogramPanel
    addViewPanel(self, HistogramPanel)


def addPowerSpectrumPanel(self, *args, **kwargs):
    """Adds a new :class:`.PowerSpectrumPanel`."""
    from fsleyes.views.powerspectrumpanel import PowerSpectrumPanel
    addViewPanel(self, PowerSpectrumPanel)


def addShellPanel(self, *args, **kwargs):
    """Adds a new :class:`.ShellPanel`."""
    from fsleyes.views.shellpanel import ShellPanel
    addViewPanel(self, ShellPanel)


def removeFocusedViewPanel(self, *args, **kwargs):
    """Removes the :class:`.ViewPanel` which currently has focus. """

    vp = self.focusedViewPanel

    if vp is not None:
        self.removeViewPanel(vp)


def _changeOverlay(self, offset):
    """Used by :func:`selectNextOverlay` and :func:`selectPreviousOverlay`.
    Changes the currently selected overlay by the given offset.
    """

    overlayList = self.overlayList

    if len(overlayList) in (0, 1):
        return

    viewPanel = self.focusedViewPanel

    if viewPanel is None: displayCtx = self     .displayCtx
    else:                 displayCtx = viewPanel.displayCtx

    cur = displayCtx.overlayOrder.index(displayCtx.selectedOverlay)
    new = displayCtx.overlayOrder[(cur + offset) % len(overlayList)]

    displayCtx.selectedOverlay = new


def selectNextOverlay(self, *args, **kwargs):
    """Increments the :attr:`.DisplayContext.selectedOverlay`. """
    _changeOverlay(self, 1)


def selectPreviousOverlay(self, *args, **kwargs):
    """Decrements the :attr:`.DisplayContext.selectedOverlay`. """
    _changeOverlay(self, -1)


def toggleOverlayVisibility(self, *args, **kwargs):
    """Shows/hides the currently selected overlay. """

    overlayList = self.overlayList

    if len(overlayList) == 0:
        return

    viewPanel = self.focusedViewPanel

    if viewPanel is None: displayCtx = self     .displayCtx
    else:                 displayCtx = viewPanel.displayCtx

    overlay = displayCtx.getSelectedOverlay()
    display = displayCtx.getDisplay(overlay)

    display.enabled = not display.enabled


def openHelp(self, *args, **kwargs):
    """Opens FSLeyes help in a web browser. """

    url = op.join(fsleyes.assetDir, 'userdoc', 'index.html')

    import fsleyes_widgets.utils.webpage as webpage

    # Show locally stored help files
    if op.exists(url):
        webpage.openFile(url)
    else:
        url = 'https://users.fmrib.ox.ac.uk/~paulmc/fsleyes/userdoc/latest/'
        webpage.openPage(url)


def setFSLDIR(self, *args, **kwargs):
    """Opens a directory dialog allowing the user to select a value for
    ``$FSLDIR``.
    """

    import wx
    from   fsl.utils.platform import platform
    import fsl.utils.settings as     settings

    fsldir = platform.fsldir

    if fsldir is None:
        fsldir = os.getcwd()

    msg = strings.titles[self, 'setFSLDIR']

    dlg = wx.DirDialog(self,
                       message=msg,
                       defaultPath=fsldir,
                       style=wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST)

    if dlg.ShowModal() != wx.ID_OK:
        return

    fsldir                 = dlg.GetPath()
    platform.fsldir        = fsldir
    settings.write('fsldir', fsldir)


def closeFSLeyes(self, *args, **kwargs):
    """Closes FSLeyes. """
    self.Close()


FSLeyesFrame.addOrthoPanel           = actions.action(addOrthoPanel)
FSLeyesFrame.addLightBoxPanel        = actions.action(addLightBoxPanel)
FSLeyesFrame.addScene3DPanel         = actions.action(addScene3DPanel)
FSLeyesFrame.addTimeSeriesPanel      = actions.action(addTimeSeriesPanel)
FSLeyesFrame.addHistogramPanel       = actions.action(addHistogramPanel)
FSLeyesFrame.addPowerSpectrumPanel   = actions.action(addPowerSpectrumPanel)
FSLeyesFrame.addShellPanel           = actions.action(addShellPanel)
FSLeyesFrame.removeFocusedViewPanel  = actions.action(removeFocusedViewPanel)
FSLeyesFrame.selectNextOverlay       = actions.action(selectNextOverlay)
FSLeyesFrame.selectPreviousOverlay   = actions.action(selectPreviousOverlay)
FSLeyesFrame.toggleOverlayVisibility = actions.action(toggleOverlayVisibility)
FSLeyesFrame.openHelp                = actions.action(openHelp)
FSLeyesFrame.setFSLDIR               = actions.action(setFSLDIR)
FSLeyesFrame.closeFSLeyes            = actions.action(closeFSLeyes)



#!/usr/bin/env python
#
# filetreemanager.py - Functions and classes used by the FileTreePanel.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains the :class:`FileTreeManager`, which is used by the
:class:`.FileTreePanel`. It also contains the :class:`OverlayManager`, which
is used by the ``FileTreeManager``.


Overview
--------


The :class:`.FileTreePanel` allows the user to navigate structured
directories, where the files and sub-directories are named according to a
:mod:`.filetree` specification. The ``FileTreePanel`` allows the user to
select which file types to display, and to restrict or re-order the files
with file tree variables.


By default, the :class:`.FileTreePanel` will display a list containing one row
for every combination of variable values; each row will contain all files for
(the selected file types) which correspond to that combination of variables.
The user may also choose to display ``<all>`` values of a specific variable on
a single row.


This module handles the mechanics of generating lists of variables and files
according to the user's settings.


In this module, and in the :mod:`.filetreepanel` module, variables which take
on a different value for each row are referred to as *varying*, and variables
for which all possible values are displaed on one row are referred to as
*fixed*.


Example
-------


For example, imagine that we have a data set with data for multiple subjects
(``sub``) and sessions (``ses``), described by this file tree::

    subj-{participant}
      [ses-{session}]
        T1w.nii.gz (T1w)
        {hemi}.gii (surface)

So for one subject and one session, we might have the following files::

    T1.nii.gz
    L.gii
    R.gii

So we have two files types (``T1`` and ``surface``), and three variables
(``sub``, ``ses``, and ``hemi``).  By default, all variables are *varying*, so
the ``FileTreePanel`` will display this data set like so (the ``x`` indicates
whether or not each file is present):


======= ======= ========= ======  ===========
``sub`` ``ses`` ``hemi``  ``T1``  ``surface``
------- ------- --------- ------  -----------
``1``   ``1``   ``L``     ``x``   ``x``
``1``   ``1``   ``R``     ``x``   ``x``
``1``   ``2``   ``L``     ``x``   ``x``
``1``   ``2``   ``R``     ``x``   ``x``
``2``   ``1``   ``L``     ``x``   ``x``
``2``   ``1``   ``R``     ``x``   ``x``
``2``   ``2``   ``L``     ``x``   ``x``
``2``   ``2``   ``R``     ``x``   ``x``
======= ======= ========= ======  ===========


However, it may make more sense to display all of the surface files together.
The user can do this by setting the ``hemi`` variable to ``<all>``, thus
changing it to a *fixed* variable. This will cause the :class:`.FileTreePanel`
to re-arrange the grid like so:


======= ======= ====== =================== ===================
``sub`` ``ses`` ``T1`` ``surface[hemi=L]`` ``surface[hemi=R]``
------- ------- ------ ------------------- -------------------
``1``   ``1``   ``x``  ``x``               ``x``
``1``   ``2``   ``x``  ``x``               ``x``
``2``   ``1``   ``x``  ``x``               ``x``
``2``   ``2``   ``x``  ``x``               ``x``
======= ======= ====== =================== ===================
"""


import                    collections
import                    logging
import functools       as ft
import itertools       as it

import fsl.utils.cache                 as cache
import fsleyes_widgets.utils.typedict  as td
import fsleyes.actions.loadoverlay     as loadoverlay


log = logging.getLogger(__name__)


FILETREE_PREFIX = '[filetree] '
"""This is a prefix added to the name of every overlay which is added to the
:class:`.OverlayList` by this module.
"""


class FileGroup(object):
    """A ``FileGroup`` represents a single row in the file tree panel list.  It
    encapsulates a set of values for all varying variables, and a set of files
    and their associated fixed variable values. These are all accessible as
    attributes called ``varyings``, ``files``, and ``fixedvars``.

    Another attribute, ``fileIDs``, contains a unique ID for each file within
    one ``FileGroup``. This ID can be used to pair up files from different
    ``FileGroup`` objects.
    """


    def __init__(self, varyings, fixed, ftypes, files):
        """Create a ``FileGroup``.

        :arg varyings: Dict of ``{ var : val }`` mappings containing the
                       varying variable values.

        :arg fixed:    List containing ``{ var : val }`` mappings, each
                       containing the fixed variable values for each file.

        :arg ftypes:   List containing the file type for each file.

        :arg files:    List of file names, the same length as ``fixedvars``.
                       Missing files are represented as ``None``.
        """

        self.varyings = varyings
        self.fixed    = fixed
        self.ftypes   = ftypes
        self.files    = files
        self.fileIDs  = []

        # Generate an ID which uniquely identifies
        # each overlay by file type and fixed
        # variable values (i.e. a unique ID for each
        # row . This will be used as
        # the overlay name
        for fname, ftype, v in zip(files, ftypes, fixed):

            # The FileGroup may contain
            # non-existent files, and
            # we don't want to load files
            # that are already in the
            # overlay list
            if fname is None:
                self.fileIDs.append(None)
                continue

            if len(v) == 0:
                fid = ftype
            else:
                fid = ftype + '[' + ','.join(
                    ['{}={}'.format(var, val)
                     for var, val in sorted(v.items())]) + ']'

            self.fileIDs.append(fid)


    def __str__(self):
        """Return a string representation of this ``FileGroup``. """
        return 'FileGroup({}, {}, {}, {})'.format(self.varyings,
                                                  self.fixed,
                                                  self.ftypes,
                                                  self.files)


    def __repr__(self):
        """Return a string representation of this ``FileGroup``. """
        return str(self)


    def __eq__(self, other):
        """Return ``True`` if this ``FileGroup`` is equal to ``other``. """
        return (self.varyings == other.varyings and
                self.fixed    == other.fixed    and
                self.ftypes   == other.ftypes   and
                self.files    == other.files)


class FileTreeManager(object):
    """The ``FileTreeManager`` class handles the generation and arranging
    of varying and fixed variables, and file types, according to a
    specification of *varying* and *fixed* variables.

    The ``FileTreeManager`` creates and uses an :class:`OverlayManager` which
    handles overlay display.
    """


    def __init__(self, overlayList, displayCtx, query):
        """Create a ``FileTreeManager``.

        :arg overlayList: The :class:`.OverlayList`

        :arg displayCtx:  The :class:`.DisplayContext` which is to manage
                          the file tree overlay display.

        :arg query:      :class:`.FileTreeQuery` instance
        """

        self.__query       = query
        self.__ovlmgr      = OverlayManager(overlayList, displayCtx)
        self.__ftypes      = None
        self.__varyings    = None
        self.__fixed       = None
        self.__varcols     = None
        self.__fixedcols   = None
        self.__filegroups  = None


    def destroy(self):
        """Must be called when this ``FileTreeManager`` is no longer needed.
        Destroys the :class:`OverlayManager` and clears references.
        """

        self.__ovlmgr.destroy()
        self.__query      = None
        self.__ovlmgr     = None
        self.__ftypes     = None
        self.__varyings   = None
        self.__fixed      = None
        self.__varcols    = None
        self.__fixedcols  = None
        self.__filegroups = None


    def update(self, ftypes, varyings, fixed):
        """Update the internal file tree grid information according to the
        given file types and variables.

        :arg ftypes:   List of file types that are to be displayed

        :arg varyings: Dict of ``{var : value}`` mappings defining the
                       varying variables.

        :arg fixed:    List of variable names defining the fixed variables.
        """

        query                 = self.query
        varyings              = prepareVaryings( query,  ftypes,   varyings)
        fixed                 = prepareFixed(    query,  ftypes,   fixed)
        varcols,    fixedcols = genColumns(      ftypes, varyings, fixed)
        filegroups            = genFileGroups(   query,  varyings, fixedcols)
        filegroups, fixedcols = filterFileGroups(filegroups,       fixedcols)

        self.__ovlmgr.update(filegroups)

        self.__ftypes     = ftypes
        self.__varyings   = varyings
        self.__fixed      = fixed
        self.__varcols    = varcols
        self.__fixedcols  = fixedcols
        self.__filegroups = filegroups


    def reorder(self, varcols):
        """Re-order the file groups according to the new varying variable
        order. The first varying variable is the slowest changing.

        :arg varcols: List of varying variable names.
        """

        if sorted(varcols) != sorted(self.varcols):
            raise ValueError('Invalid varying columns: {}'.format(varcols))

        filegroups = self.filegroups

        def cmp(g1, g2):

            g1 = filegroups[g1]
            g2 = filegroups[g2]

            for col in varcols:
                v1 = g1.varyings[col]
                v2 = g2.varyings[col]
                if   v1 == v2:   continue
                elif v1 is None: return  1
                elif v2 is None: return -1
                elif v1 > v2:    return  1
                elif v1 < v2:    return -1
            return 0

        idxs = list(range(len(filegroups)))
        idxs = sorted(idxs, key=ft.cmp_to_key(cmp))

        # Update the internal manager state
        # so the new order is preserved
        filegroups = [filegroups[i] for i in idxs]
        varyings   = collections.OrderedDict()
        varcols    = list(varcols)

        for v in varcols:
            varyings[v] = self.varyings[v]

        self.__ovlmgr.update(filegroups)

        self.__filegroups = filegroups
        self.__varyings   = varyings
        self.__varcols    = varcols


    def show(self, filegroup, callback=None):
        """Show the overlays associated with a :class:`FileGroup`.

        All arguments are passed through to the :meth:`OverlayManager.show` method.
        """
        self.__ovlmgr.show(filegroup, callback)


    @property
    def query(self):
        """Returns the :class:`.FileTreeQuery` object used by this
        ``FileTreeManager``.
        """
        return self.__query


    @property
    def ftypes(self):
        """Returns a list of all file types to be displayed. """
        return self.__ftypes


    @property
    def varyings(self):
        """Returns a dict of ``{ var : [value] }`` mappings, containing every
        possible value for each varying variable.
        """
        return self.__varyings


    @property
    def fixed(self):
        """Returns a dict of ``{ ftype : { var : [value] } }`` mappings which,
        for each file type, contains a dictionary of all fixed variables and
        their possible values.
        """
        return self.__fixed


    @property
    def varcols(self):
        """Returns a list of varying variable names to be used as columns for
        the varying variables.
        """
        return self.__varcols


    @property
    def fixedcols(self):
        """Returns a list of tuples, with each tuple containing:
          - A file type
          - A dict of ``{var : value}`` mappings, containing
            fixed variable values

        Each tuple represents a column for a combination of file type and
        fixed variable values.
        """
        return self.__fixedcols


    @property
    def filegroups(self):
        """Returns a list containing all of the ``FileGroup`` objects. Each
        ``FileGroup`` represents one row in the file tree grid.
        """
        return self.__filegroups


def prepareVaryings(query, ftypes, varyings):
    """Called by :meth:`FileTreeManager.update`. Prepares a dictionary which
    contains all possible values for each varying variable.

    :arg query:    :class:`.FileTreeQuery` object

    :arg ftypes:   List of file types to be displayed.

    :arg varyings: Dict of ``{ var : value }`` mappings. A value of ``'*'``
                   indicates that all possible values for this variable
                   should be used.

    :returns:      A dict of ``{ var : [value] }`` mappings, containing
                   every possible value for each varying variable.
    """

    allvars = query.variables()

    # Force a constsient ordering
    # of the varying variables
    _varyings = collections.OrderedDict()
    for var in sorted(varyings.keys()):
        _varyings[var] = varyings[var]
    varyings = _varyings

    # Expand the varying dict so that
    # it contains { var : [value] }
    # mappings - '*' is replaced with
    # a list of all possible values,
    # and a scalar value is replaced
    # with a list containing just that
    # value.
    for var, val in list(varyings.items()):

        # This variable is not relevant for
        # any of the specified file types.
        if not any([var in query.variables(ft) for ft in ftypes]):
            varyings.pop(var)
            continue

        elif val == '*': varyings[var] = allvars[var]
        else:            varyings[var] = [val]

    return varyings


def prepareFixed(query, ftypes, fixed):
    """Called by :meth:`.FileTreeManager.update`. Prepares a dictionary
    which contains all possible values for each fixed variable, and for
    each file type.

    :arg query:  :class:`.FileTreeQuery` object

    :arg ftypes: List of file types to be displayed

    :arg fixed:  List of fixed variables

    :returns:    A dict of ``{ ftype : { var : [value] } }`` mappings
                 which, for each file type, contains a dictionary of
                 all fixed variables and their possible values.
    """

    allvars = query.variables()

    # Create a dict for each file type
    # containing { var : [value] }
    # mappings for all fixed variables
    # which are relevant to that file
    # type.
    _fixed = {}
    for ftype in ftypes:
        ftvars        = query.variables(ftype)
        _fixed[ftype] = {}
        for var in fixed:
            if var in ftvars:
                _fixed[ftype][var] = allvars[var]
    fixed = _fixed

    return fixed


def genColumns(ftypes, varyings, fixed):
    """Determines all columns which need to be present in a file tree grid
    for the given file types, varying and fixed variables.

    :arg ftypes:   List of file types to be displayed

    :arg varyings: Dict of ``{ var : [value} }`` mappings, containing all
                   varying variables and their possible values (see
                   :func:`prepareVaryings`).

    :arg fixed:    Dict of ``{ ftype : { var : [value] } }`` mappings
                   which, for each file type, contains a dictionary of
                   all fixed variables and their possible values.

    :returns:      Two lists which, combined, represent all columns to be
                   displayed in the file tree grid:

                    - A list of varying variable names
                    - A list of tuples, with each tuple containing:

                      - A file type
                      - A dict of ``{var : value}`` mappings, containing
                        fixed variable values
    """

    varcols   = [var for var, vals in varyings.items() if len(vals) > 1]
    fixedcols = []

    for ftype in ftypes:

        ftvars    = fixed[ftype]
        ftvarprod = list(it.product(*[vals for vals in ftvars.values()]))

        for ftvals in ftvarprod:
            ftvals = {var : val for var, val in zip(ftvars, ftvals)}
            fixedcols.append((ftype, ftvals))

    return varcols, fixedcols


def genFileGroups(query, varyings, fixed):
    """Generates a list of :class:`FileGroup` objects, each representing one
    row in a grid defined by the given set of varying and fixed variables.

    :arg query:    :class:`.FileTreeQuery` object

    :arg varyings: Dict of ``{ var : [value} }`` mappings, containing all
                   varying variables and their possible values (see
                   :func:`prepareVaryings`).

    :arg fixed:    List of tuples of ``(ftype, { var : value })`` mappings,
                   which each contain a file type and set of fixed variables
                   corresponding to one column in the grid.

    :returns:      A list of ``FileGroup`` objects.
    """

    if len(varyings) == 0:
        return []

    # Build a list of file groups - each
    # file group represents a group of
    # files to be displayed together,
    # corresponding to a combination of
    # varying values
    filegroups = []

    # loop through all possible
    # combinations of varying values
    for vals in it.product(*varyings.values()):

        groupVars   = {var : val for var, val in zip(varyings.keys(), vals)}
        groupFtypes = []
        groupFixed  = []
        groupFiles  = []

        for ftype, ftvars in fixed:

            fname = query.query(ftype, **groupVars, **ftvars)

            # There should only be one file for
            # each combination of varying+fixed
            # values
            if len(fname) == 1: fname = fname[0].filename
            else:               fname = None

            groupFtypes.append(ftype)
            groupFixed .append(ftvars)
            groupFiles .append(fname)

        grp = FileGroup(groupVars, groupFixed, groupFtypes, groupFiles)
        filegroups.append(grp)

    return filegroups


def filterFileGroups(filegroups, fixedcols):
    """Filters out empty, duplicate and redundant rows, and empty columns,
    from ``filegroups``

    :arg filegroups: List of :class:`FileGroup` objects.
    :arg fixedcols:  List of ``(ftype, { var : value })`` mappings
    :returns:        A tuple containing the filtered ``filegroups`` and
                     ``fixedcols``
    """

    # TODO optimise this whole thing

    dropcols = set()

    # Remove duplicate/redundant rows
    for i in range(len(filegroups)):

        if i in dropcols:
            continue

        grpi   = filegroups[i]
        ifiles = [f for f in grpi.files if f is not None]

        # drop empty rows
        if len(ifiles) == 0:
            dropcols.add(i)
            continue

        for j in range(i + 1, len(filegroups)):

            if j in dropcols:
                continue

            grpj   = filegroups[j]
            jfiles = [f for f in grpj.files if f is not None]

            # Group i contains all the files
            # of group j - we can drop group j
            # (this will also cause j to be
            # dropped if it is empty)
            if all([n in ifiles for n in jfiles]):
                dropcols.add(j)

            # Group j contains all the files
            # in group i - we can drop group i
            elif all([n in jfiles for n in ifiles]):
                dropcols.add(i)
                break

    filegroups = [g for i, g in enumerate(filegroups) if i not in dropcols]

    # Count the number of present files in
    # each column, and drop empty columns
    if len(filegroups) > 0: ncolumns = len(filegroups[0].files)
    else:                   ncolumns = 0

    colcounts = {i : 0 for i in range(ncolumns)}
    for grp in filegroups:
        for i, fname in enumerate(grp.files):
            if fname is not None:
                colcounts[i] += 1

    keepcols = [idx for idx, count in colcounts.items() if count > 0]

    if len(keepcols) > 0 and len(keepcols) < ncolumns:
        fixedcols = [fixedcols[i] for i in keepcols]
        for grp in filegroups:
            grp.fileIDs = [grp.fileIDs[i] for i in keepcols]
            grp.files   = [grp.files[  i] for i in keepcols]
            grp.ftypes  = [grp.ftypes[ i] for i in keepcols]
            grp.fixed   = [grp.fixed[  i] for i in keepcols]

    return filegroups, fixedcols


class OverlayManager(object):
    """The ``OverlayManager`` is used by the :class:`FileTreeManager`. It
    manages the mechanics of displaying overlays associated with the file tree.

    The :meth:`update` method is used to tell the ``OverlayManager`` about the
    currently displayed list of :class:`FileGroup` objects. The :meth:`show`
    method is used to show the overlays in a specific ``FileGroup``.

    Whenever the :meth:`show` method is called, the overlays from any
    previously displayed ``FileGroup`` are "swapped" out for the overlays in
    the new ``FileGroup``. The display properties for matching pairs of
    overlays are preserved as best as possible.
    """


    def __init__(self, overlayList, displayCtx):
        """Create an ``OverlayManager``

        :arg overlayList: The :class:`.OverlayList`

        :arg displayCtx:  The :class:`.DisplayContext` which is to manage
                          the file tree overlay display.
        """

        self.__overlayList = overlayList
        self.__displayCtx  = displayCtx

        # Loaded overlays are cached to
        # reduce swap time if they are
        # re-shown.
        self.__cache = cache.Cache(maxsize=50, lru=True)

        # The current list of file
        # groups - set in update
        self.__filegroups = None

        # When we swap one group out
        # for another, we preserve
        # the overlay ordering.
        self.__order = None

        # When we swap one group out
        # for another, we preserve
        # overlay display/opts
        # property values.
        self.__propVals = None


    def destroy(self):
        """Must be called when this ``OverlayManager`` is no longer needed.
        Clears references.
        """
        self.__cache.clear()

        self.__overlayList = None
        self.__displayCtx  = None
        self.__cache       = None
        self.__filegroups  = None
        self.__order       = None
        self.__propVals    = None


    def update(self, filegroups):
        """Must be called when the list of ``FileGroup`` objects has changed,
        either due to them being re-ordered or completely changed.
        """
        self.__filegroups = filegroups
        self.__order      = []
        self.__propVals   = {}


    def show(self, filegroup, callback=None):
        """Show the overlays associated with the given :class:`FileGroup`.

        Any overlays which were previously displayed are removed, and replaced
        with the overlays associated with the new group.

        :arg filegroup: ``FileGroup`` to show
        :arg callback:  Optional function which will be called when the
                        overlays have been shown. Will not be called if no new
                        overlays are to be shown.
        """

        # Force an error if a file
        # group which is not in our
        # known list is passed in.
        self.__filegroups.index(filegroup)

        displayCtx = self.__displayCtx

        # Gather all of the files
        # to be loaded, using their
        # unique file ID (unique
        # within one file group)
        new = {}
        for fname, fid in zip(filegroup.files, filegroup.fileIDs):

            # The FileGroup may contain
            # non-existent files
            if fname is None:
                continue

            new[FILETREE_PREFIX + fid] = fname

        # Get refs to the existing
        # filetree overlays - we will
        # replace the existing ones
        # with the new ones.
        old = collections.OrderedDict()
        for ovl in displayCtx.getOrderedOverlays():
            if ovl.name.startswith(FILETREE_PREFIX):
                old[ovl.name] = ovl

        # Save the display settings
        # of the old overlays - we'll
        # apply them to the new ones
        for fid, ovl in old.items():
            propVals = self.__propVals.get(fid, {})
            propVals.update(getProperties(ovl, displayCtx))
            self.__propVals[fid] = propVals

        # Save the ordering of the old
        # overlays so we can preserve
        # it in the new overlays. The
        # length check is to take into
        # account missing files -
        # file groups with more present
        # files should take precedence
        # over file groups with more
        # missing files.
        if len(old) >= len(self.__order):
            self.__order = list(old.keys())

        # Set the ordering of the new
        # overlays based on the old ones
        order = self.__order
        fids  = [f for f in order if f     in new] + \
                [f for f in new   if f not in order]
        new   = collections.OrderedDict([(f, new[f]) for f in fids])

        if len(new) > 0:
            self.__load(new, old, callback)


    def __load(self, new, old, callback=None):
        """Called by :meth:`show`. Loads the files specified in ``new``, then
        passes them (along with the ``old``) to the :meth:`__show` method.

        :arg new:      Dict of ``{fileid : file}`` mappings, containing the
                       files to load.

        :arg old:      Dict of ``{fileid : overlay}`` mappings, containing the
                       existing overlays which will be replaced with the new
                       ones.

        :arg callback: No-args function which will be called after the new
                       overlays have been loaded.

        """

        # Remove any files that are
        # in the cache, and don't
        # need to be re-loaded
        cache     = self.__cache
        toload    = [(fid, f) for fid, f in new.items() if fid not in cache]
        loadfids  = [f[0] for f in toload]
        loadfiles = [f[1] for f in toload]

        def onLoad(ovlidxs, ovls):

            # Gather the overlays to be shown,
            # either from the ones that were
            # just loaded, or from the cache.
            toshow = collections.OrderedDict()

            for fid, fname in new.items():

                # Is this overlay in the cache?
                if fname in cache:
                    toshow[fid] = cache[fname]

                # Or has it just been loaded?
                else:
                    idx         = loadfids.index(fid)
                    toshow[fid] = ovls[idx]
                    cache.put(fname, ovls[idx])

            self.__show(toshow, old)

            if callback is not None:
                callback()

        loadoverlay.loadOverlays(loadfiles, onLoad=onLoad)


    def __show(self, new, old):
        """Adds the given ``new`` overlays to the :class:`.OverlayList`. The
        display properties of any ``old`` overlays with the same ID are copied
        over to the new ones.

        All existing overlays which were previously added are removed.
        """

        overlayList = self.__overlayList
        displayCtx  = self.__displayCtx

        # Drop old overlays that
        # are being (re-)addd
        old = collections.OrderedDict(
            [(fid, ovl) for fid, ovl in old.items()
             if ovl not in new.values()])

        # Drop new overlays that
        # are already in the list,
        # but keep a ref to the
        # original, in case any
        # properties refer to them
        allnew = new
        new    = collections.OrderedDict(
            [(fid, ovl) for fid, ovl in new.items()
             if ovl not in overlayList])

        for fid, ovl in new.items(): log.debug('Adding %s: %s', fid, ovl)
        for      ovl in old:         log.debug('Removing %s',   ovl)

        # We set the name of each new overlay
        # to its file ID. If the user changes
        # the overlay name, it will no longer
        # be tracked
        for fid, ovl in new.items():
            ovl.name = fid

        # Copy the display settings across from
        # the old overlays to the new ones. The
        # old ones were saved in the show method,
        # but we need to rearrange them a little
        # before passing them to the overlayList.
        propVals = collections.defaultdict(dict)

        for fid, ovl in new.items():
            for name, val in self.__propVals.get(fid, {}).items():

                # If the value is a ToReplace
                # object, it has has been
                # REPLACEd by the getProperties
                # function - see REPLACE and
                # getProperties.
                if isinstance(val, ToReplace):
                    if val.value in allnew: val = allnew[val.value]
                    else:                   val = None

                propVals[name][ovl] = val

        # remove the old overlays from any
        # overlay groups, otherwise property
        # syncing might screw things up
        old = list(old.values())
        for ovl in old:
            for group in displayCtx.overlayGroups:
                group.removeOverlay(ovl)

        # Insert the new overlays into the list.
        overlayList.extend(new.values(), **propVals)

        # preserve the display space if
        # it was set to a filegroup image
        dspace = displayCtx.displaySpace
        if dspace in old and dspace.name in new:
            displayCtx.displaySpace = new[dspace.name]

        # Remove all of the
        # old filetree overlays
        idxs           = range(len(overlayList))
        idxs           = [i for i in idxs if overlayList[i] not in old]
        overlayList[:] = [overlayList[i] for i in idxs]



REPLACE = td.TypeDict({
    'MeshOpts'    : ['refImage'],
    'VolumeOpts'  : ['clipImage'],
    'VectorOpts'  : ['colourImage', 'modulateImage', 'clipImage']
})
"""This dict contains :class:`.DisplayOpts` properties which refer to other
images, and which need to be explicitly handled when the
:class:`OverlayManager` swaps a group of overlays in for another.
"""


SKIP = td.TypeDict({
    'DisplayOpts' : ['bounds'],
    'NiftiOpts'   : ['transform'],
    'MeshOpts'    : ['vertexSet', 'vertexData', 'vertexDataIndex'],
})
"""This dict contains :class:`.DisplayOpts` properties which are not copied
when the :class:`OverlayManager` swaps a group of overlays in for an existing
group.
"""


class ToReplace(object):
    """Placeholder type used by the :func:`getProperties` function when a
    property value is in the :attr:`REPLACE` dictionary, and needs to be
    explicitly handled by the :class:`OverlayManager`.
    """
    def __init__(self, value):
        self.value = value


def getProperties(ovl, displayCtx):
    """Retrieves the :class:`.Display` and :class:`DisplayOpts` properties
    for the given overlay, applying the rules defined by the :attr:`REPLACE`
    and :attr:`SKIP` dictionaries.

    :arg ovl:        An overlay
    :arg displayCtx: The :class:`.DisplayContext` managing the overlay display.
    :returns:        a dict of ``{ name : value}`` mappings.
    """

    propVals = {}
    display  = displayCtx.getDisplay(ovl)
    opts     = displayCtx.getOpts(   ovl)
    skip     = list(it.chain(*SKIP   .get(opts, [], allhits=True)))
    replace  = list(it.chain(*REPLACE.get(opts, [], allhits=True)))

    for propName in display.getAllProperties()[0]:
        propVals[propName] = getattr(display, propName)

    for propName in opts.getAllProperties()[0]:

        if propName in skip:
            continue

        value = getattr(opts, propName)

        if value is None:
            continue

        if (propName in replace) and (value is not None):
            value = ToReplace(value.name)

        propVals[propName] = value

    return propVals



#!/usr/bin/env python
#
# powerspectrumtoolbar.py - The PowerSpectrumToolBar class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`PowerSpectrumToolBar`, a toolbar for use
with a :class:`.PowerSpectrumPanel`.
"""


import fsleyes_props    as props

import fsleyes.icons    as icons
import fsleyes.tooltips as tooltips
import fsleyes.actions  as actions


from . import plottoolbar


class PowerSpectrumToolBar(plottoolbar.PlotToolBar):
    """The ``PowerSpectrumToolBar`` is a toolbar for use with a
    :class:`.PowerSpectrumPanel`. It extends :class:`.PlotToolBar`,
    and adds a few controls specific to the :class:`.PoweSpectrumPanel`.
    """

    def __init__(self, parent, overlayList, displayCtx, frame, psPanel):
        """Create a ``PowerSpectrumToolBar``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        :arg psPanel:     The :class:`.PowerSpectrumPanel` instance.
        """

        plottoolbar.PlotToolBar.__init__(
            self, parent, overlayList, displayCtx, frame, psPanel)

        togControl = actions.ToggleActionButton(
            'togglePowerSpectrumControl',
            actionKwargs={'floatPane' : True},
            icon=[icons.findImageFile('spannerHighlight24'),
                  icons.findImageFile('spanner24')],
            tooltip=tooltips.actions[psPanel, 'togglePowerSpectrumControl'])

        togList = actions.ToggleActionButton(
            'togglePlotList',
            actionKwargs={'floatPane' : True},
            icon=[icons.findImageFile('listHighlight24'),
                  icons.findImageFile('list24')],
            tooltip=tooltips.actions[psPanel, 'togglePlotList'])

        togControl = props.buildGUI(self, psPanel, togControl)
        togList    = props.buildGUI(self, psPanel, togList)

        self.InsertTools([togControl, togList], 0)

        nav = [togControl, togList] + self.getCommonNavOrder()

        self.setNavOrder(nav)


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``PowerSpectrumToolBar`` is only intended to be added to
        :class:`.PowerSpectrumPanel` views.
        """
        from fsleyes.views.powerspectrumpanel import PowerSpectrumPanel
        return [PowerSpectrumPanel]
